@@ -2,7 +2,7 @@
  *    hlist.h
  *
  *    history list
- *    Copyright (C) 2002-2003  George M. Tzoumas
+ *    Copyright (C) 2002-2003,2011  George M. Tzoumas
  *
  *    This program is free software; you can redistribute it and/or modify
  *    it under the terms of the GNU General Public License as published by
@@ -38,7 +38,7 @@ typedef struct {
 void hlist_init(hlist_t *h, int limit);
 void hline_free(hline_t *l);
 void hlist_free(hlist_t *h);
-hline_t *hlist_find(hlist_t *h, char *s);
-hline_t *hlist_add(hlist_t *h, char *s, int len, int attr);
+hline_t *hlist_find(hlist_t *h, const char *s);
+hline_t *hlist_add(hlist_t *h, const char *s, int len, int attr);
 
 #endif
