@@ -2,7 +2,7 @@
  *    msn.c
  *
  *    gtmess - MSN Messenger client
- *    Copyright (C) 2002-2004  George M. Tzoumas
+ *    Copyright (C) 2002-2011  George M. Tzoumas
  *
  *    This program is free software; you can redistribute it and/or modify
  *    it under the terms of the GNU General Public License as published by
@@ -19,17 +19,22 @@
  *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
-#include<unistd.h>
-#include<stdlib.h>
-#include<stdio.h>
-#include<string.h>
-#include<iconv.h>
+#include <unistd.h>
+#include <stdlib.h>
+#include <stdio.h>
+#include <string.h>
+#include <iconv.h>
 
-#include"../config.h"
+#include <sys/errno.h>
 
-#include"md5.h"
-#include"msn.h"
-#include"../inty/inty.h"
+#include "../config.h"
+
+#include "md5.h"
+#include "msn.h"
+#include "../inty/inty.h"
+#include "util.h"
+
+#include <assert.h>
 
 char *msn_stat_name[] = {
     "Offline", "Appear Offline", "Online", "Idle",
@@ -40,18 +45,41 @@ char *msn_stat_com[] = { "FLN", "HDN", "NLN", "IDL",
 char msn_stat_char[10] = "  >-=*_#&?";
 
 iconv_t msn_ic[2];
-        
+
+msn_group_t msn_group_0 = { "* individuals *", "", NULL };
+
 /******* contact lists *******/
 
-msn_contact_t *msn_clist_find(msn_clist_t *q, char *login)
+msn_contact_t *msn_clist_find(msn_clist_t *q, unsigned char lf, char *login)
 {
     msn_contact_t *p;
 
     for (p = q->head; p != NULL; p = p->next)
-        if (strcmp(p->login, login) == 0) break;
+        if ((p->lflags & lf) == lf && strcmp(p->login, login) == 0) break;
     return p;
 }
 
+msn_contact_t *msn_clist_findu(msn_clist_t *q, unsigned char lf, char *uuid)
+{
+    msn_contact_t *p;
+	
+    for (p = q->head; p != NULL; p = p->next)
+        if ((p->lflags & lf) == lf && strcmp(p->uuid, uuid) == 0) break;
+    return p;
+}
+
+int msn_clist_count(msn_clist_t *q, unsigned char lf)
+{
+    msn_contact_t *p;
+    int c = 0;
+
+    for (p = q->head; p != NULL; p = p->next) if ((p->lflags & lf) == lf) {
+        if (p->lflags & msn_FL) c += p->groups; else c++;
+    }
+    return c;
+}
+
+/* FL only *
 int msn_clist_gcount(msn_clist_t *q, char *login)
 {
     msn_contact_t *p;
@@ -60,65 +88,100 @@ int msn_clist_gcount(msn_clist_t *q, char *login)
     for (p = q->head; p != NULL; p = p->next)
         if (strcmp(p->login, login) == 0) c++;
     return c;
-}
+} */
 
+/* FL only */
 /* return val:
     -1 = contact not in list
-     0 = contact in list, but up to date
+     0 = contact in list, but up-to-date
      1 = contact updated
 */
-int msn_clist_update(msn_clist_t *q, char *login,
-                     char *nick, int status, int blocked, int tm_last_char)
+int msn_clist_update(msn_clist_t *q, unsigned char lf, char *loginu,
+                     char *nick, int status, int tm_last_char, int byuuid)
 {
     msn_contact_t *p;
-    int r = 0, f = -1;
+    int r = -1;
+    char *s;
 
-    for (p = q->head; p !=NULL; p = p->next)
-        if (strcmp(p->login, login) == 0) {
-            f++;
-            if (nick != NULL) {
-                if (strcmp(p->nick, nick) != 0) p->dirty = 1;
-                strcpy(p->nick, nick);
-                r = 1;
+    for (p = q->head; p !=NULL; p = p->next) {
+        if (byuuid) s = p->uuid; else s = p->login;
+        if (((p->lflags & lf) == lf) && strcmp(s, loginu) == 0) {
+            r = 0;
+            if (nick != NULL && strcmp(p->nick, nick) != 0) {
+                    p->dirty = 1;
+                    Strcpy(p->nick, nick, SML);
+                    r = 1;
             }
-            if (status >= 0) p->status = status, r = 1;
-            if (blocked >= 0) p->blocked = blocked, r = 1;
+            if (status >= 0 && p->status != status) p->status = status, r = 1;
             if (tm_last_char >= 0) p->tm_last_char = tm_last_char, r = 1;
+            break;
         }
-    if (f == -1) return -1;
-    else return r;
+    }
+    return r;
+}
+
+int msn_clist_psm_update(msn_clist_t *q, char *login, char *psm)
+{
+    msn_contact_t *p;
+    int r = -1;
+
+    for (p = q->head; p !=NULL; p = p->next) {
+        if (strcmp(p->login, login) == 0) {
+            r = 0;
+            if (psm != NULL) {
+                if (strcmp(p->psm, psm) != 0) {
+                    Strcpy(p->psm, psm, SML);
+                    r = 1;
+                }
+            }
+            break;
+        }
+    }
+    return r;
 }
 
 void msn_contact_free(msn_contact_t *q)
 {
+    if (q == NULL) return;
     if (q->next != NULL) msn_contact_free(q->next);
+    if (q->gid != NULL) {
+        int i;
+        assert(q->gidsize > 0);
+        for (i = 0; i < q->groups; i++)	free(q->gid[i]);
+        free(q->gid);
+    }
     free(q);
 }
 
 void msn_clist_free(msn_clist_t *q)
 {
     if (q->head != NULL) msn_contact_free(q->head);
-    q->head = NULL;
     q->count = 0;
+    q->head = NULL;
 }
 
-int msn_clist_rem_grp(msn_clist_t *q, msn_clist_t *rl, int gid)
+int msn_clist_rem_zombies(msn_clist_t *q)
 {
-    msn_contact_t *p, *rev;
+    msn_contact_t *p;
     msn_contact_t *last, *next;
     int c = 0;
 
     p = q->head;
     last = NULL;
     while (1) {
-        while (p != NULL && (p->gid != gid)) {
+        while (p != NULL && p->lflags != 0) {
             last = p;
             p = p->next;
         }
         if (p != NULL) {
             next = p->next;
-            if (rl != NULL && (rev = msn_clist_find(rl, p->login)) != NULL)
-                rev->gid--;
+            assert(p->groups == 0);
+            /*if (p->gid != NULL) {
+                int i;
+                for (i = 0; i < p->groups; i++) free(p->gid[i]);
+                free(p->gid);
+            }*/
+            if (p->gidsize) free(p->gid);
             free(p); c++;
             if (last != NULL) last->next = next; else q->head = next;
             p = next;
@@ -128,26 +191,68 @@ int msn_clist_rem_grp(msn_clist_t *q, msn_clist_t *rl, int gid)
     return c;
 }
 
-void msn_clist_rem(msn_clist_t *q, char *login, int gid)
+int msn_clist_rem_grp(msn_clist_t *q, char *gid)
 {
-    msn_contact_t *p = q->head;
-    msn_contact_t *last = NULL, *next;
+    msn_contact_t *p;
+    int i, c = 0;
 
-    while (p != NULL && (strcmp(p->login, login) != 0 || (gid >= 0 && p->gid != gid))) {
-        /* de morgan ! */
-        /* ! ((p == NULL) || (strcmp(p->login, login == 0) && (gid < 0 || p->gid == gid))) */
-        last = p;
-        p = p->next;
-    }
-    if (p != NULL) {
-        next = p->next;
-        free(p);
-        if (last != NULL) last->next = next; else q->head = next;
-        q->count--;
+    for (p = q->head; p != NULL; p = p->next) {
+        /* linear search, no problem since #groups small */
+        for (i = 0; i < p->groups; i++) if (strcmp(p->gid[i], gid) == 0) {
+			free(p->gid[i]);
+            if (i < p->groups-1) memmove(&p->gid[i], &p->gid[i+1], (p->groups-i-1)*sizeof(char *));
+            p->groups--;
+            c++;
+            break;
+        }
     }
+    return c;
 }
 
-int msn_clist_load(msn_clist_t *q, FILE *f, int count)
+void msn_clist_rem(msn_clist_t *q, unsigned char lf, char *loginu, char *gid)
+{
+    msn_contact_t *p;
+    char *s;
+    int i;
+
+    for (p = q->head; p != NULL; p = p->next) {
+        if (lf == msn_FL) s = p->uuid; else s = p->login;
+        if (strcmp(s, loginu) != 0) continue;
+        if ((p->lflags & lf) != lf) continue;
+        
+        if (lf == msn_FL) {
+            /* linear search, no problem since #groups small */
+            if (gid != NULL) {
+                /* remove contact from given group ONLY */
+                if (p->groups) {
+                    i = 0;
+                    while (i < p->groups)
+                    if (strcmp(p->gid[i], gid) == 0) {
+                        free(p->gid[i]);
+                        if (i < p->groups-1) memmove(&p->gid[i], &p->gid[i+1], (p->groups-i-1)*sizeof(char *));
+                        p->groups--;
+                        break;
+                    } else i++;
+                }
+            } else {
+                /* remove contact from ALL groups and list */
+                if (p->gid) {
+                    for (i = 0; i < p->groups; i++) free(p->gid[i]);
+                    free(p->gid);
+                    p->gid = NULL;
+                    p->groups = 0;
+                    p->gidsize = 0;
+                }
+                p->lflags &= ~lf;
+            }
+        } else p->lflags &= ~lf;
+        
+        break; /* weird style, isn't it? :-) */
+    }
+    msn_clist_rem_zombies(q);
+}
+
+/* int msn_clist_load(msn_clist_t *q, FILE *f, int count)
 {
     msn_contact_t t, *p, *cur, *last;
     int c = 0;
@@ -158,7 +263,7 @@ int msn_clist_load(msn_clist_t *q, FILE *f, int count)
         p = (msn_contact_t *) malloc(sizeof(msn_contact_t));
         if (p == NULL) return -1;
         t.status = MS_FLN;
-        /* sorted insertion */
+        /* sorted insertion *
         cur = q->head;
         last = NULL;
         memcpy(p, &t, sizeof(t));
@@ -172,9 +277,9 @@ int msn_clist_load(msn_clist_t *q, FILE *f, int count)
         c++;
     }
     return c;
-}
+} */
 
-int msn_clist_save(msn_clist_t *q, FILE *f)
+/* int msn_clist_save(msn_clist_t *q, FILE *f)
 {
     msn_contact_t *p;
     int c = 0;
@@ -183,38 +288,86 @@ int msn_clist_save(msn_clist_t *q, FILE *f)
         c++;
     }
     return c;
-}
+} */
 
-msn_contact_t *msn_clist_add(msn_clist_t *q, int gid, char *login, char *nick)
+
+/* gid == NULL: no gids will be stored (contact gid stays NULL) */
+msn_contact_t *msn_clist_add(msn_clist_t *q, unsigned char lf, char *login, char *nick, char *gid, char *uuid)
 {
-    msn_contact_t *p, *cur, *last;
+    msn_contact_t *p;
+
+    if ((p = msn_clist_find(q, 0, login)) != NULL) {
+        if (gid != NULL && gid[0] && (lf & msn_FL)) {
+            if (!msn_contact_belongs(p, gid)) {
+                if (p->groups >= p->gidsize) {
+                    if (p->gidsize == 0) p->gidsize = 4; else p->gidsize *= 2;
+                    p->gid = (char **) realloc(p->gid, p->gidsize * sizeof(char *));
+                }
+                p->gid[p->groups++] = strdup(gid);
+            }
+        }
+        p->lflags |= lf;
+        return p;
+    }
 
     p = (msn_contact_t *) calloc(1, sizeof(msn_contact_t));
     if (p == NULL) return NULL;
-    p->gid = gid;
-    p->blocked = 0;
+    p->lflags = lf;
+    p->groups = 0;
+
+    if (gid != NULL && gid[0]) {
+        p->gidsize = 4;
+        p->gid = (char **) malloc(p->gidsize * sizeof(char *));
+        if (p->gid == NULL) {
+            free(p);
+            return NULL;
+        }
+        p->gid[p->groups++] = strdup(gid);
+    } else {
+        p->gid = NULL;
+        p->gidsize = 0;
+    }
+
     p->tm_last_char = 0;
-    strcpy(p->login, login);
-    strcpy(p->nick, nick);
+    Strcpy(p->login, login, SML);
+    if (nick != NULL && nick[0]) Strcpy(p->nick, nick, SML);
+    else Strcpy(p->nick, p->login, SML);
+    if (uuid != NULL && uuid[0]) Strcpy(p->uuid, uuid, SNL);
+    else p->uuid[0] = 0;
+    p->psm[0] = 0;
     p->dirty = 0;
     p->status = MS_FLN;
+    p->notify = 1;
+    p->ignored = 0;
 
-    /* sorted insertion */
-    cur = q->head;
-    last = NULL;
-    while (cur != NULL && p->gid > cur->gid) last = cur, cur = cur->next;
-    p->next = cur;
-    if (last != NULL) {
-        last->next = p;
-    } else q->head = p;
-
-/*    p->next = q->head;
-    q->head = p;*/
+    p->next = q->head;
+    q->head = p;
     q->count++;
     return p;
 }
 
-void msn_clist_cpy(msn_clist_t *dest, msn_clist_t *src)
+/* if gid == NULL, then returns 1 when contact is GROUPLESS! */
+int msn_contact_belongs(const msn_contact_t *q, char *gid)
+{
+    int i;
+    if (gid == NULL) return q->groups == 0;
+    for (i = 0; i < q->groups; i++)
+        if (strcmp(q->gid[i], gid) == 0) return 1;
+    return 0;
+}
+
+void msn_contact_cpy(msn_contact_t *dest, msn_contact_t *src)
+{
+    memcpy(dest, src, sizeof(msn_contact_t));
+    if (dest->gid != NULL) {
+        int i;
+        dest->gid = (char **) malloc(dest->gidsize * sizeof(char *));
+		for (i = 0; i < dest->groups; i++) dest->gid[i] = strdup(src->gid[i]);
+    }
+    dest->next = NULL;
+}
+
+void msn_clist_cpy(msn_clist_t *dest, msn_clist_t *src, unsigned char lf)
 {
     msn_contact_t *p, *q;
 
@@ -222,9 +375,10 @@ void msn_clist_cpy(msn_clist_t *dest, msn_clist_t *src)
     dest->head = NULL;
     q = dest->head;
     for (p = src->head; p != NULL; p = p->next) {
+        if ((p->lflags & lf) != lf) continue;
         q = (msn_contact_t *) malloc(sizeof(msn_contact_t));
         if (q == NULL) return;
-        memcpy(q, p, sizeof(msn_contact_t));
+        msn_contact_cpy(q, p);
         q->next = dest->head;
         dest->head = q;
         dest->count++;
@@ -233,21 +387,21 @@ void msn_clist_cpy(msn_clist_t *dest, msn_clist_t *src)
 
 /******* group lists *******/
 
-msn_group_t *msn_glist_find(msn_glist_t *q, int gid)
+msn_group_t *msn_glist_find(msn_glist_t *q, char *gid)
 {
     msn_group_t *p;
 
     for (p = q->head; p != NULL; p = p->next)
-        if (p->gid == gid) break;
+        if (strcmp(p->gid, gid) == 0) break;
     return p;
 }
 
-char *msn_glist_findn(msn_glist_t *q, int gid)
+char *msn_glist_findn(msn_glist_t *q, char *gid)
 {
     msn_group_t *p;
 
     for (p = q->head; p != NULL; p = p->next)
-        if (p->gid == gid) return p->name;
+        if (strcmp(p->gid, gid) == 0) return p->name;
     return "<unknown>";
 }
 
@@ -264,22 +418,22 @@ void msn_glist_free(msn_glist_t *q)
     q->count = 0;
 }
 
-void msn_glist_ren(msn_glist_t *q, int gid, char *newname)
+void msn_glist_ren(msn_glist_t *q, char *gid, char *newname)
 {
     msn_group_t *p = q->head;
 
-    while (p != NULL && p->gid != gid) p = p->next;
+    while (p != NULL && strcmp(p->gid, gid) != 0) p = p->next;
     if (p != NULL) {
-        strcpy(p->name, newname);
+        Strcpy(p->name, newname, SML);
     }
 }
 
-void msn_glist_rem(msn_glist_t *q, int gid)
+void msn_glist_rem(msn_glist_t *q, char *gid)
 {
     msn_group_t *p = q->head;
     msn_group_t *last = NULL, *next;
 
-    while (p != NULL && p->gid != gid) {
+    while (p != NULL && strcmp(p->gid, gid) != 0) {
         last = p;
         p = p->next;
     }
@@ -291,7 +445,7 @@ void msn_glist_rem(msn_glist_t *q, int gid)
     }
 }
 
-int msn_glist_load(msn_glist_t *q, FILE *f, int count)
+/* int msn_glist_load(msn_glist_t *q, FILE *f, int count)
 {
     msn_group_t t, *p, *cur, *last;
     int c = 0;
@@ -301,7 +455,7 @@ int msn_glist_load(msn_glist_t *q, FILE *f, int count)
         if (fread(&t, sizeof(t), 1, f) != 1) return -1;
         p = (msn_group_t *) malloc(sizeof(msn_group_t));
         if (p == NULL) return -1;
-        /* sorted insertion */
+        /* sorted insertion *
         cur = q->head;
         last = NULL;
         memcpy(p, &t, sizeof(t));
@@ -326,28 +480,30 @@ int msn_glist_save(msn_glist_t *q, FILE *f)
         c++;
     }
     return c;
-}
+} */
 
-void msn_glist_add(msn_glist_t *q, int gid, char *name)
+msn_group_t *msn_glist_add(msn_glist_t *q, char *gid, char *name)
 {
     msn_group_t *p, *cur, *last;
 
     p = (msn_group_t *) calloc(1, sizeof(msn_group_t));
-    if (p == NULL) return;
-    p->gid = gid;
-    strcpy(p->name, name);
+    if (p == NULL) return NULL;
+    Strcpy(p->gid, gid, SNL);
+    Strcpy(p->name, name, SML);
 
     cur = q->head;
     last = NULL;
-    while (cur != NULL && p->gid > cur->gid) last = cur, cur = cur->next;
+    /* insertion sort by name */
+    while (cur != NULL && strcmp(p->name, cur->name) > 0) last = cur, cur = cur->next;
     p->next = cur;
     if (last != NULL) {
         last->next = p;
     } else q->head = p;
     q->count++;
+    return p;
 }
 
-void msn_glist_cpy(msn_glist_t *dest, msn_glist_t *src)
+void msn_glist_cpy(msn_glist_t *dest, msn_glist_t *src, msn_contact_t *ref)
 {
     msn_group_t *p, *q;
 
@@ -355,6 +511,13 @@ void msn_glist_cpy(msn_glist_t *dest, msn_glist_t *src)
     dest->head = NULL;
     q = dest->head;
     for (p = src->head; p != NULL; p = p->next) {
+        /* reference contact: copy only groups contact belongs to */
+        if (ref != NULL) {
+            int i;
+            for (i = 0; i < ref->groups; i++)
+                if (strcmp(ref->gid[i], p->gid) == 0) break;
+            if (i == ref->groups) continue;
+        }
         q = (msn_group_t *) malloc(sizeof(msn_group_t));
         if (q == NULL) return;
         memcpy(q, p, sizeof(msn_group_t));
@@ -366,7 +529,7 @@ void msn_glist_cpy(msn_glist_t *dest, msn_glist_t *src)
 
 /******* string utility routines *******/
 
-void md5hex(char *src, char *hex_output)
+void md5hex(const char *src, char *hex_output)
 {
     md5_state_t state;
     md5_byte_t digest[16];
@@ -379,33 +542,7 @@ void md5hex(char *src, char *hex_output)
         sprintf(hex_output + di * 2, "%02x", digest[di]);
 }
 
-void str2url(char *src, char *dest)
-{
-    static char unreserved[256] = {
-      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
-      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
-      0, 1, 0, 0, 1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, /* !$'()*+,-. */
-      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, /* 0123456789 */
-      0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* ABCDEFGHIJKLMNOP */
-      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, /* QRSTUVWXYZ_*/
-      0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* abcdefghijklmnop */
-      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, /* qrstuvwxyz */
-      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
-      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
-      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
-      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
-      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
-      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
-      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
-      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
-
-    for (; *src; src++)
-        if (unreserved[(unsigned char) *src]) *dest++ = *src;
-        else sprintf(dest, "%%%02X", (unsigned char) *src), dest += 3;
-    *dest = 0;
-}
-
-void utf8decode(iconv_t ic, char *src, char *dest)
+void utf8decode(iconv_t ic, const char *src, char *dest)
 {
     size_t ibl, obl;
     
@@ -430,7 +567,7 @@ void utf8decode(iconv_t ic, char *src, char *dest)
     }
 }
 
-void utf8encode(iconv_t ic, char *src, char *dest, int obl)
+void utf8encode(iconv_t ic, const char *src, char *dest, size_t obl)
 {
     size_t ibl;
     
@@ -455,8 +592,56 @@ void utf8encode(iconv_t ic, char *src, char *dest, int obl)
     }
 }
 
+void str2url(const char *src, char *dest)
+{
+    static char unreserved[256] = {
+      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+      0, 1, 0, 0, 1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, /* !$'()*+,-. */
+      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, /* 0123456789 */
+      0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* ABCDEFGHIJKLMNOP */
+      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, /* QRSTUVWXYZ_*/
+      0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /* abcdefghijklmnop */
+      1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, /* qrstuvwxyz */
+      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+      0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
 
-void url2str(char *src, char *dest)
+    char textubuf[2*SXL], *textu;
+    const char *s;
+    int alloc = 0;
+    size_t size, len;
+
+    len = strlen(src);
+    if (msn_ic[1] != (iconv_t) -1) {
+        if (6*len < 2*SXL) {
+            textu = textubuf;
+            alloc = 0;
+            size = 2*SXL;
+        } else {
+            size = 6*len;
+            textu = (char *) malloc(size+1);
+            alloc = 1;
+        }
+        utf8encode(msn_ic[1], src, textu, size);
+        s = textu;
+    } else {
+        s = src;
+    }
+    
+    for (; *s; s++)
+        if (unreserved[(unsigned char) *s]) *dest++ = *s;
+        else sprintf(dest, "%%%02X", (unsigned char) *s), dest += 3;
+    *dest = 0;
+    if (alloc && textu != NULL) free(textu);
+}
+
+void url2str(const char *src, char *dest)
 {
     unsigned int c;
     char *tmp, *t;
@@ -475,11 +660,47 @@ void url2str(char *src, char *dest)
     free(tmp);
 }
 
-int is3(char *a, char *b)
+int is3(const char *a, const char *b)
 {
     return a[0] == b[0] && a[1] == b[1] && a[2] == b[2];
 }
 
+
+/* buf contains server data, starting from "X=..." (just after tid)
+   if login/nick/uuid are not NULL, then parsed data are stored there;
+   if no data is found, then strings become EMPTY!
+   returns a pointer to the first character after the last token parsed (usually ' ') */
+char *msn_parse_contact_data(char *buf, char *login, char *nick, char *uuid)
+{
+    char *cur, t;
+    char tmp[SML];
+
+    if (login == NULL) login = &tmp[0];
+    if (nick == NULL) nick = &tmp[0];
+    if (uuid == NULL) uuid = &tmp[0];
+    login[0] = nick[0] = uuid[0] = 0;
+    cur = strchr(buf, '=');
+    while (1) {
+        char t;
+        tmp[0] = 0;
+        if (cur == NULL) return buf;
+        t = *(cur-1);
+        if (t == 'N') {
+            sscanf(cur+1, "%s", login);
+            cur += strlen(login) + 1; /* +1 for '=' */
+        } else if (t == 'F') {
+            sscanf(cur+1, "%s", tmp);
+            cur += strlen(tmp) + 1;
+            if (nick != &tmp[0]) url2str(tmp, nick);
+        } else if (t == 'C') {
+            sscanf(cur+1, "%s", uuid);
+            cur += strlen(uuid) + 1;
+        }
+        buf = cur;
+        cur = strchr(buf, '=');
+    }
+}
+
 /******* errors *******/
 
 char *msn_error_str(int err)
@@ -512,6 +733,10 @@ char *msn_error_str(int err)
         case 301: return "too many hits to FND";
         case 302: return "not logged in";
 
+        case 402:
+        case 403: return "error accessing contact list";
+        case 420: return "invalid account permissions";
+
         case 500: return "internal server error";
         case 501: return "database server error";
         case 502: return "command disabled";
@@ -553,6 +778,8 @@ char *msn_error_str(int err)
         case 922: return "server too busy";
         case 923: return "kids passport without parental consent";
         case 924: return "passport account not yet verified";
+        case 928: return "bad ticket";
+        case 931: return "account not on this server";
 
         default: return "<unknown server error>";
     }
@@ -589,6 +816,11 @@ msn_stat_t msn_stat_id(char *s)
 
 /******* protocol *******/
 
+int writestr(int fd, char *s)
+{
+    return write(fd, s, strlen(s));
+}
+
 /* ping server */
 int msn_png(int fd)
 {
@@ -601,82 +833,194 @@ int msn_out(int fd)
     return write(fd, "OUT\r\n", 5);
 }
 
-/* reply to server challenge */
-int msn_qry(int fd, unsigned int tid, char *hash)
+int is_little_endian(void)
 {
-    char reply[SNL], hex[SNL];
+  int w = 1;
+  return *((char *) &w);
+}
 
-    strcat(hash, "Q1P7W2E4J9R8U3S5");
-    md5hex(hash, hex);
-    sprintf(reply, "QRY %u msmsgs@msnmsgr.com 32\r\n%s", tid, hex);
-    return write(fd, reply, strlen(reply));
+unsigned int convert_endianess(unsigned int i)
+{
+    unsigned int r;
+    r  = (i & 0x000000FF) << 0x18;
+    r |= (i & 0x0000FF00) << 0x08;
+    r |= (i & 0x00FF0000) >> 0x08;
+    r |= (i & 0xFF000000) >> 0x18;
+    return r;
+}
+
+/* adapted from original code by Tibor Billes */
+int msn_chl(const char *challenge, const char *product_id, char *hash)
+{
+    char md5[36];
+    unsigned int i[4], *j;
+    int loop;
+    size_t size;
+
+    /* product_key must be equal to or greater than product_id */
+    static const char *product_key = "YMM8C_H7KCQ2S_KL";
+    char *chg;
+    long long int low = 0, high = 0;
+    int hash_[4], h1, h2, h3, h4;
+
+    size = strlen(challenge) + strlen(product_key) + 8;
+    chg = (char *) Malloc(size);
+    if (chg == NULL) return -1;
+    j = (unsigned int *) chg;
+
+    strcpy(chg, challenge); strcat(chg, product_key);
+    md5hex(chg, md5);
+
+    sscanf(md5, "%8x%8x%8x%8x", &i[0], &i[1], &i[2], &i[3]);
+
+    /* the string md5hex returned is big endian */
+    i[0] = convert_endianess(i[0]) & 0x7FFFFFFF;
+    i[1] = convert_endianess(i[1]) & 0x7FFFFFFF;
+    i[2] = convert_endianess(i[2]) & 0x7FFFFFFF;
+    i[3] = convert_endianess(i[3]) & 0x7FFFFFFF;
+
+    strcpy(chg, challenge); strcat(chg, product_id);
+    for (loop = strlen(chg); loop % 8; loop++) chg[loop] = '0';
+    chg[loop+1] = 0;
+
+    if (!is_little_endian()) {
+        for (loop = 0; loop < strlen(chg); loop += 4) {
+            /* `/4' looks more beautiful and should get converted to >> 2
+                by the compiler */
+            j[loop/4] = convert_endianess(j[loop/4]);
+        }
+    }
+
+    for (loop = 0; loop < strlen(chg)/4-1; loop += 2) {
+        long long int temp = j[loop];
+        temp = (temp * 0x0E79A9C1) % 0x7FFFFFFF;
+        temp += high;
+        temp = i[0] * temp + i[1];
+        temp = temp % 0x7FFFFFFF;
+        high = j[loop+1];
+        high = (high + temp) % 0x7FFFFFFF;
+        high = high * i[2] + i[3];
+        high = high % 0x7FFFFFFF;
+
+        low += high + temp;
+    }
+    high = (high + i[1]) % 0x7FFFFFFF;
+    low = (low + i[3]) % 0x7FFFFFFF;
+
+    sscanf(md5, "%8x%8x%8x%8x", &h1, &h2, &h3, &h4);
+
+    high = convert_endianess(high);
+    low = convert_endianess(low);
+    hash_[0] = h1^high;
+    hash_[1] = h2^low;
+    hash_[2] = h3^high;
+    hash_[3] = h4^low;
+
+    sprintf(hash, "%08x%08x%08x%08x", hash_[0], hash_[1], hash_[2], hash_[3]);
+    free(chg);
+    return 0;
+}
+
+/* reply to server challenge */
+int msn_qry(int fd, unsigned int tid, const char *challenge)
+{
+    char reply[SNL], hash[33];
+    static const char *product_id  = "PROD0090YUAUV{2B";
+
+    msn_chl(challenge, product_id, hash);
+    sprintf(reply, "QRY %u %s 32\r\n%s", tid, product_id, hash);
+    return writestr(fd, reply);
 }
 
 /* change user status */
 int msn_chg(int fd, unsigned int tid, msn_stat_t status)
 {
-    char s[SNL];
+    char s[SML];
 
     sprintf(s, "CHG %u %s\r\n", tid, msn_stat_com[status]);
-    return write(fd, s, strlen(s));
+    return writestr(fd, s);
+}
+
+/* set personal message */
+int msn_uux(int fd, unsigned int tid, const char *psm)
+{
+    char reply[SML*3], s[SML*2], psmu[SML];
+    str2url(psm, psmu);
+    sprintf(s, " <Data><PSM>%s</PSM><CurrentMedia></CurrentMedia></Data>", psmu);
+    sprintf(reply, "UUX %u %zu\r\n%s", tid, strlen(s), s);
+    return writestr(fd, reply);
 }
 
 /* sync contact lists */
 int msn_syn(int fd, unsigned int tid, unsigned int ver)
 {
-    char s[SNL];
+    char s[SML];
 
-    sprintf(s, "SYN %u %u\r\n", tid, ver);
-    return write(fd, s, strlen(s));
+    sprintf(s, "SYN %u %u 0\r\n", tid, ver);
+    return writestr(fd, s);
 }
 
 /* change group name */
-int msn_reg(int fd, unsigned int tid, int gid, char *name)
+int msn_reg(int fd, unsigned int tid, char *gid, char *name)
 {
     char s[SML], urln[SML];
     
     str2url(name, urln);
-    sprintf(s, "REG %u %d %s 0\r\n", tid, gid, urln);
-    return write(fd, s, strlen(s));
+    sprintf(s, "REG %u %s %s\r\n", tid, gid, urln);
+    return writestr(fd, s);
 }
 
-/* change name */
-int msn_rea(int fd, unsigned int tid, char *login, char *name)
+/* change contact name */
+int msn_sbp(int fd, unsigned int tid, char *uuid, char *name)
 {
     char s[SML], urln[SML];
    
     str2url(name, urln);
-    sprintf(s, "REA %u %s %s\r\n", tid, login, urln);
-    return write(fd, s, strlen(s));
+    sprintf(s, "SBP %u %s MFN %s\r\n", tid, uuid, urln);
+    return writestr(fd, s);
+}
+
+/* change MY name */
+int msn_prp(int fd, unsigned int tid, char *name)
+{
+    char s[SML], urln[SML];
+
+    str2url(name, urln);
+    sprintf(s, "PRP %u MFN %s\r\n", tid, urln);
+    return writestr(fd, s);
 }
 
 /* remove from contact list */
-int msn_rem(int fd, unsigned int tid, char list, char *login, int gid)
+int msn_rem(int fd, unsigned int tid, char list, char *loginu, char *gid)
 {
     char s[SML];
     
-    if (list == 'F') sprintf(s, "REM %u %cL %s %d\r\n", tid, list, login, gid);
-    else sprintf(s, "REM %u %cL %s\r\n", tid, list, login);
-    return write(fd, s, strlen(s));
+    if (list == 'F') {
+        if (gid != NULL) sprintf(s, "REM %u FL %s %s\r\n", tid, loginu, gid);
+        else sprintf(s, "REM %u FL %s\r\n", tid, loginu);
+    } else sprintf(s, "REM %u %cL %s\r\n", tid, list, loginu);
+    return writestr(fd, s);
 }
 
 /* add to contact list */
-int msn_add(int fd, unsigned int tid, char list, char *login, int gid)
+int msn_adc(int fd, unsigned int tid, char list, char *loginu, char *gid)
 {
     char s[SML];
     
-    if (list == 'F') sprintf(s, "ADD %u %cL %s %s %d\r\n", tid, list, login, login, gid);
-    else sprintf(s, "ADD %u %cL %s %s\r\n", tid, list, login, login);
-    return write(fd, s, strlen(s));
+    if (list == 'F') {
+        if (gid == NULL || !gid[0]) sprintf(s, "ADC %u FL N=%s F=%s\r\n", tid, loginu, loginu);
+        else sprintf(s, "ADC %u FL C=%s %s\r\n", tid, loginu, gid);
+    } else sprintf(s, "ADC %u %cL N=%s\r\n", tid, list, loginu);
+    return writestr(fd, s);
 }
 
 /* remove group */
-int msn_rmg(int fd, unsigned int tid, int gid)
+int msn_rmg(int fd, unsigned int tid, char *gid)
 {
     char s[SML];
     
-    sprintf(s, "RMG %u %d\r\n", tid, gid);
-    return write(fd, s, strlen(s));
+    sprintf(s, "RMG %u %s\r\n", tid, gid);
+    return writestr(fd, s);
 }
 
 /* add group */
@@ -685,8 +1029,8 @@ int msn_adg(int fd, unsigned int tid, char *group)
     char s[SML], urln[SML];
     
     str2url(group, urln);
-    sprintf(s, "ADG %u %s 0\r\n",tid, urln);
-    return write(fd, s, strlen(s));
+    sprintf(s, "ADG %u %s\r\n",tid, urln);
+    return writestr(fd, s);
 }
 
 /* reverse list prompting: (A)lways/(N)ever */
@@ -695,7 +1039,7 @@ int msn_gtc(int fd, unsigned int tid, char c)
     char s[SML];
     
     sprintf(s, "GTC %u %c\r\n", tid, c);
-    return write(fd, s, strlen(s));
+    return writestr(fd, s);
 }
 
 /* all other users: (B)locked/(A)llowed */
@@ -704,7 +1048,7 @@ int msn_blp(int fd, unsigned int tid, char c)
     char s[SML];
     
     sprintf(s, "BLP %u %cL\r\n", tid, c);
-    return write(fd, s, strlen(s));
+    return writestr(fd, s);
 }
 
 int msn_cvr(int fd, unsigned int tid, char *cvr, char *login)
@@ -714,7 +1058,7 @@ int msn_cvr(int fd, unsigned int tid, char *cvr, char *login)
         sprintf(s, "CVR %u %s\r\n", tid, cvr);
     else
         sprintf(s, "CVR %u %s %s\r\n", tid, cvr, login);
-    return write(fd, s, strlen(s));
+    return writestr(fd, s);
 }
 
 /* starts the procedure to log in a notification/dispatch server 
@@ -732,20 +1076,40 @@ int msn_login_init(int fd, unsigned int tid, char *login, char *cvr, char *dest)
     int err;
     char s[SXL], rep[SXL], tmp[SML];
     
-    sprintf(s, "VER %u MSNP9 CVR0\r\n", tid++);
-    if (write(fd, s, strlen(s)) < 0) return -1;
-
-    if (readlnt(fd, rep, SXL, SOCKET_TIMEOUT) == NULL) return -2;
+/*    sprintf(s, "VER %u MSNP12 CVR0\r\n", tid++);*/
+    sprintf(s, "VER %u MSNP12\r\n", tid++);
+    if (writestr(fd, s) < 0) {
+        err = errno;
+        close(fd);
+        errno = err;
+        return -1;
+    }
+    if (readlnt(fd, rep, SXL, SOCKET_TIMEOUT) == NULL) {
+        err = errno;
+        close(fd);
+        errno = err;
+        return -2;
+    }
     if (sscanf(rep, "VER %*u %s", tmp) != 1 || 
-        strcmp(tmp, "MSNP9") != 0) {
+        strcmp(tmp, "MSNP12") != 0) {
             close(fd);
             if (sscanf(rep, "%d", &err) == 1) return err;
             return -3;
     }
 
-    if (msn_cvr(fd, tid++, cvr, login) < 0) return -1;
+    if (msn_cvr(fd, tid++, cvr, login) < 0) {
+        err = errno;
+        close(fd);
+        errno = err;
+        return -1;
+    }
     
-    if (readlnt(fd, rep, SXL, SOCKET_TIMEOUT) == NULL) return -2;
+    if (readlnt(fd, rep, SXL, SOCKET_TIMEOUT) == NULL) {
+        err = errno;
+        close(fd);
+        errno = err;
+        return -2;
+    }
     if (sscanf(rep, "%s", tmp) != 1 || strcmp(tmp, "CVR") != 0) {
         close(fd);
         if (sscanf(rep, "%d", &err) == 1) return err;
@@ -753,11 +1117,21 @@ int msn_login_init(int fd, unsigned int tid, char *login, char *cvr, char *dest)
     }
 
     sprintf(s, "USR %u TWN I %s\r\n", tid++, login);
-    if (write(fd, s, strlen(s)) < 0) return -1;
+    if (writestr(fd, s) < 0) {
+        err = errno;
+        close(fd);
+        errno = err;
+        return -1;
+    }
 
-    if (readlnt(fd, rep, SXL, SOCKET_TIMEOUT) == NULL) return -2;
+    if (readlnt(fd, rep, SXL, SOCKET_TIMEOUT) == NULL) {
+        err = errno;
+        close(fd);
+        errno = err;
+        return -2;
+    }
     if (sscanf(rep, "XFR %*u %*s %s", tmp) == 1) {
-        if (dest != NULL) strcpy(dest, tmp);
+        if (dest != NULL) Strcpy(dest, tmp, SML);
         close(fd);
         return 1;
     }
@@ -770,23 +1144,33 @@ int msn_login_init(int fd, unsigned int tid, char *login, char *cvr, char *dest)
 }
 
 /* finalizes the login procedure to a notification server using TWN authentication
-it retrieves the user's nickname into `nick'
     uses 1 tid
     returns 
     0 on success */
-int msn_login_twn(int fd, unsigned int tid, char *ticket, char *nick)
+int msn_login_twn(int fd, unsigned int tid, char *ticket)
 {
     int err;
-    char s[SXL], rep[SXL];
+    char s[SXL];
+    int foo;
     
     sprintf(s, "USR %u TWN S %s\r\n", tid, ticket);
-    if (write(fd, s, strlen(s)) < 0) return -1;
+    if (writestr(fd, s) < 0) {
+        err = errno;
+        close(fd);
+        errno = err;
+        return -1;
+    }
     
-    if (readlnt(fd, rep, SXL, SOCKET_TIMEOUT) == NULL) return -2;
-    if (sscanf(rep, "USR %*u OK %*s %s", s) != 1) {
-        if (sscanf(rep, "%d", &err) == 1) return err;
+    if (readlnt(fd, s, SXL, SOCKET_TIMEOUT) == NULL) {
+        err = errno;
+        close(fd);
+        errno = err;
+        return -2;
+    }
+    if (sscanf(s, "USR %*u OK %*s %d", &foo) != 1) {
+        if (sscanf(s, "%d", &err) == 1) return err;
         return -3;
-    } else if (nick != NULL) url2str(s, nick);
+    }
     
     return 0;
 }
@@ -799,8 +1183,8 @@ int msn_msg_typenotif(int fd, unsigned int tid, char *user)
     sprintf(msg, "MIME-Version: 1.0\r\n"
                  "Content-Type: text/x-msmsgscontrol\r\n"
                  "TypingUser: %s\r\n\r\n\r\n", user);
-    sprintf(s, "MSG %u U %d\r\n%s", tid, strlen(msg), msg);
-    return write(fd, s, strlen(s));
+    sprintf(s, "MSG %u U %zu\r\n%s", tid, strlen(msg), msg);
+    return writestr(fd, s);
 }
 
 /* 
@@ -810,7 +1194,7 @@ commands(args):
     GTMESS = tell everybody you are using gtmess, too
     SAY(text) = send <text> in a way that only gtmess users can see
 */
-int msn_msg_gtmess(int fd, unsigned int tid, char *cmd, char *args)
+int msn_msg_gtmess(int fd, unsigned int tid, const char *cmd, const char *args)
 {
     char s[2*SXL+6*48], msg[2*SXL+5*48], argsu[2*SXL];
 
@@ -819,21 +1203,35 @@ int msn_msg_gtmess(int fd, unsigned int tid, char *cmd, char *args)
                  "Content-Type: text/x-gtmesscontrol\r\n"
                  "Command: %s\r\n"
                  "Args: %s\r\n\r\n", cmd, argsu);
-    sprintf(s, "MSG %u U %d\r\n%s", tid, strlen(msg), msg);
-    return write(fd, s, strlen(s));
+    sprintf(s, "MSG %u U %zu\r\n%s", tid, strlen(msg), msg);
+    return writestr(fd, s);
 }
 
 /* send text message */
 int msn_msg_text(int fd, unsigned int tid, char *text)
 {
-    char s[2*SXL+5*48], msg[2*SXL+4*48], textu[2*SXL];
+    char s[SNL], textubuf[2*SXL], *textu;
+    int alloc;
+    size_t size, len;
 
-    utf8encode(msn_ic[1], text, textu, 2*SXL);
-    sprintf(msg, "MIME-Version: 1.0\r\n"
-                 "Content-Type: text/plain; charset=UTF-8\r\n\r\n"
-                 "%s", textu);
-    sprintf(s, "MSG %u N %d\r\n%s", tid, strlen(msg), msg);
-    return write(fd, s, strlen(s));
+    len = strlen(text);
+    if (6*len < 2*SXL) {
+        textu = textubuf;
+        alloc = 0;
+        size = 2*SXL;
+    } else {
+        size = 6*len;
+        textu = (char *) malloc(size+1);
+        alloc = 1;
+    }
+    utf8encode(msn_ic[1], text, textu, size);
+    sprintf(s, "MSG %u N %zu\r\n", tid, strlen(textu)+62);
+    writestr(fd, s);
+    writestr(fd, "MIME-Version: 1.0\r\n"
+                 "Content-Type: text/plain; charset=UTF-8\r\n\r\n");    /* len = 62 */
+    writestr(fd, textu);
+    if (alloc && textu != NULL) free(textu);
+    return len;
 }
 
 /* send file invitation */
@@ -852,8 +1250,8 @@ int msn_msg_finvite(int fd, unsigned int tid, unsigned int cookie,
                  "Invitation-Cookie: %d\r\n"
                  "Application-File: %s\r\n"
                  "Application-FileSize: %u\r\n\r\n", cookie, fnameu, size);
-    sprintf(s, "MSG %u N %d\r\n%s", tid, strlen(msg), msg);
-    return write(fd, s, strlen(s));
+    sprintf(s, "MSG %u N %zu\r\n%s", tid, strlen(msg), msg);
+    return writestr(fd, s);
 }
 
 /* send info for accepted (outgoing) file invitation */
@@ -871,8 +1269,8 @@ int msn_msg_accept2(int fd, unsigned int tid, unsigned int cookie,
                  "AuthCookie: %u\r\n"
                  "Launch-Application: FALSE\r\n"
                  "Request-Data: IP-Address:\r\n\r\n", cookie, ip, port, auth_cookie);
-    sprintf(s, "MSG %u N %d\r\n%s", tid, strlen(msg), msg);
-    return write(fd, s, strlen(s));
+    sprintf(s, "MSG %u N %zu\r\n%s", tid, strlen(msg), msg);
+    return writestr(fd, s);
 }
 
 /* accept (file) invitation */
@@ -886,8 +1284,8 @@ int msn_msg_accept(int fd, unsigned int tid, unsigned int cookie)
                  "Invitation-Cookie: %u\r\n"
                  "Launch-Application: FALSE\r\n"
                  "Request-Data: IP-Address:\r\n\r\n", cookie);
-    sprintf(s, "MSG %u N %d\r\n%s", tid, strlen(msg), msg);
-    return write(fd, s, strlen(s));
+    sprintf(s, "MSG %u N %zu\r\n%s", tid, strlen(msg), msg);
+    return writestr(fd, s);
 }
 
 /* cancel invitation */
@@ -900,8 +1298,8 @@ int msn_msg_cancel(int fd, unsigned int tid, unsigned int cookie, char *code)
                  "Invitation-Command: CANCEL\r\n"
                  "Invitation-Cookie: %u\r\n"
                  "Cancel-Code: %s\r\n\r\n", cookie, code);
-    sprintf(s, "MSG %u N %d\r\n%s", tid, strlen(msg), msg);
-    return write(fd, s, strlen(s));
+    sprintf(s, "MSG %u N %zu\r\n%s", tid, strlen(msg), msg);
+    return writestr(fd, s);
 }
 
 /* call (invite) a user to a switchboard */
@@ -910,5 +1308,5 @@ int msn_cal(int fd, unsigned int tid, char *user)
     char s[SML];
 
     sprintf(s, "CAL %u %s\r\n", tid, user);
-    return write(fd, s, strlen(s));
+    return writestr(fd, s);
 }
