@@ -2,7 +2,7 @@
  *    pass.h
  *
  *    gtmess - MSN Messenger client
- *    Copyright (C) 2002-2004  George M. Tzoumas
+ *    Copyright (C) 2002-2007  George M. Tzoumas
  *
  *    This program is free software; you can redistribute it and/or modify
  *    it under the terms of the GNU General Public License as published by
@@ -25,6 +25,6 @@
 void pass_init();
 void pass_done();
 char *get_login_server(char *dest);
-int get_ticket(char *server, char *login, char *pass, char *param, char *dest);
+int get_ticket(char *server, char *login, char *pass, char *param, char *dest, size_t n);
 
 #endif
