@@ -2,7 +2,7 @@
  *    sboard.c
  *
  *    gtmess - MSN Messenger client
- *    Copyright (C) 2002-2005  George M. Tzoumas
+ *    Copyright (C) 2002-2011  George M. Tzoumas
  *
  *    This program is free software; you can redistribute it and/or modify
  *    it under the terms of the GNU General Public License as published by
@@ -19,29 +19,38 @@
  *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
-#include<curses.h>
-#include<pthread.h>
-#include<stdlib.h>
-#include<iconv.h>
-#include<errno.h>
-#include<sys/stat.h>
-#include<string.h>
-#include<time.h>
+#include <curses.h>
+#include <pthread.h>
+#include <stdlib.h>
+#include <iconv.h>
+#include <errno.h>
+#include <sys/stat.h>
+#include <string.h>
+#include <time.h>
 
-#include"sboard.h"
-#include"xfer.h"
-#include"unotif.h"
-#include"../inty/inty.h"
-#include"gtmess.h"
+#include "sboard.h"
+#include "xfer.h"
+#include "unotif.h"
+#include "../inty/inty.h"
+#include "gtmess.h"
+#include "hash_tbl.h"
+
+#ifndef ECONNRESET
+#define ECONNRESET 0
+#endif
 
 pthread_mutex_t SX; /* switchboard mutex */
 msn_sblist_t SL; /* switchboard */
 pthread_key_t key_sboard;
+
 xqueue_t q_sb_req; /* new-sb requests */
+/*  type 0: stores SB pointer (to reuse existing SB)
+    type 1: stores LOGIN (to invite immediately) */
 
 time_t dlg_time = 0;
 
 extern char MyIP[];
+extern char *help_str[];
 
 void sboard_leave_all(int panic)
 {
@@ -84,7 +93,16 @@ int sboard_leave()
 int _sboard_kill()
 {
     int r;
+    time_t now;
+
+    time(&now);
     LOCK(&SX);
+    if (Config.safe_msg >= 0 && SL.count > 0 && 
+            difftime(now, SL.head->real_dlg_now) <= Config.safe_msg) {
+        UNLOCK(&SX);
+        flash();
+        return -1; /* too soon */
+    }
     if ((r = msn_sblist_rem(&SL))) draw_sb(SL.head, 1);
     UNLOCK(&SX);
     return r;
@@ -100,7 +118,7 @@ int sboard_kill()
         pthread_cond_timedwait(&SL.out_cond, &SX, &timeout);
         UNLOCK(&SX);
     }
-    return _sboard_kill();;
+    return _sboard_kill();
 }
 
 
@@ -144,6 +162,26 @@ void sboard_next_pending()
     UNLOCK(&SX);
 }
 
+int says_ok(msn_sboard_t *sb, char *nick)
+{
+    time_t now;
+    
+    if (strcmp(sb->last_nick, nick) != 0) {
+        time(&sb->last_spoken);
+        strcpy(sb->last_nick, nick);
+        return 1;
+    }
+    time(&now);
+    if (difftime(now, sb->last_spoken) >= Config.skip_says) {
+        sb->last_spoken = now;
+        strcpy(sb->last_nick, nick);
+        return 1;
+    } else {
+        sb->last_spoken = now;
+        return 0;
+    }
+}
+
 void sb_blah()
 {
     msn_sboard_t *sb;
@@ -152,10 +190,9 @@ void sb_blah()
     LOCK(&msn.lock);
     LOCK(&SX);
     sb = SL.head;
-    if (sb != NULL && strlen(sb->input) > 0) {
+    if (sb != NULL && strlen((char *) sb->EB.text) > 0) {
         if (sb->sfd > -1) {
-            strcpy(s, sb->input);
-            sp = s;
+            sp = (char *) sb->EB.text;
             pthread_setspecific(key_sboard, (void *) sb);
             if (sp[0] == '/') {
                 if (sp[1] == '/') sp++;
@@ -164,12 +201,14 @@ void sb_blah()
                     /* special command */
                     char cmd[SXL], args[SXL];
 
-                    eb_history_add(&SL.head->EB, SL.head->EB.text, strlen(SL.head->EB.text));
-                    eb_settext(&SL.head->EB, ZS);
-                    draw_sbebox(SL.head, 1);
+                    eb_history_add(&SL.head->EB, (char *) SL.head->EB.text, strlen((char *) SL.head->EB.text));
 
                     cmd[0] = args[0] = 0;
                     sscanf(sp, "%s %[^\n]", cmd, args);
+                    
+                    eb_settext(&SL.head->EB, ZS);
+                    draw_sbebox(SL.head, 1);
+                    
                     if (strcmp(cmd, "/flash") == 0)
                     /* easter egg! */
                         flash();
@@ -184,16 +223,19 @@ void sb_blah()
                         msn_msg_typenotif(sb->sfd, sb->tid++, args);
                     } else if (strcmp(cmd, "/beep") == 0) {
                         msn_msg_gtmess(sb->sfd, sb->tid++, "BEEP", ZS);
+                    } else if (strcmp(cmd, "/help") == 0) {
+                        msg(C_DBG, "%s\n", help_str[0]);
+                        msg(C_DBG, "%s\n", help_str[1]);
                     } else if (strcmp(cmd, "/gtmess") == 0) {
                         msn_msg_gtmess(sb->sfd, sb->tid++, "GTMESS", ZS);
                     } else if (strcmp(cmd, "/msg") == 0) {
                         msn_msg_gtmess(sb->sfd, sb->tid++, "MSG", args);
-                        msg(C_DBG, "%s says:\n", getnick(msn.login, msn.nick, Config.aliases));
-                        msg(C_MSG, "%s\n", args);
+                        msg(C_DBG, "%s says:\n", msn.nick);
+                        msgn(C_MSG, strlen(args)+2, "%s\n", args);
                     } else if (strcmp(cmd, "/dlg") == 0) {
                         msn_msg_gtmess(sb->sfd, sb->tid++, "DLG", args);
-                        dlg(C_DBG, "%s says:\n", getnick(msn.login, msn.nick, Config.aliases));
-                        dlg(C_MSG, "%s\n", args);
+                        dlg(C_DBG, "%s says:\n", msn.nick);
+                        dlgn(C_MSG, strlen(args)+2, "%s\n", args);
                     } else if (strcmp(cmd, "/file") == 0) {
                         struct stat st;
                         unsigned int cookie;
@@ -202,6 +244,7 @@ void sb_blah()
                                 cookie = lrand48();
                                 xfl_add_file(&XL, XS_INVITE, msn.login, ZS, 0, cookie, sb, args, st.st_size);
                                 msn_msg_finvite(sb->sfd, sb->tid++, cookie, args, st.st_size);
+                                draw_xfer(1);
                             } else msg(C_ERR, "Could not stat file '%s'\n", args);
                         } else msg(C_ERR, "Cannot send file: MSNFTP server is not running\n");
                     } else msg(C_ERR, "Invalid command\n");
@@ -226,9 +269,10 @@ void sb_blah()
                 }
             }
             msn_msg_text(sb->sfd, sb->tid++, sp);
-            dlg(C_DBG, "%s says:\n", getnick(msn.login, msn.nick, Config.aliases));
-            dlg(C_MSG, "%s\n", sp);
-            eb_history_add(&SL.head->EB, SL.head->EB.text, strlen(SL.head->EB.text));
+            if (says_ok(sb, msn.login))
+                dlg(C_DBG, "%s says:\n", msn.nick);
+            dlgn(C_MSG, strlen(sp)+2, "%s\n", sp);
+            eb_history_add(&SL.head->EB, (char *) SL.head->EB.text, strlen((char *) SL.head->EB.text));
             eb_settext(&SL.head->EB, ZS);
             draw_sbebox(SL.head, 1);
         } else  if (sb->multi == 0) { /* reconnect only if not multiuser */
@@ -245,23 +289,24 @@ void sb_blah()
     UNLOCK(&msn.lock);
 }
 
-void sb_type(int c)
+int sb_type(int c)
 {
     msn_sboard_t *sb;
     int command;
     time_t now;
+    int res;
 
     LOCK(&msn.lock);
     LOCK(&SX);
     sb = SL.head;
     if (sb != NULL) {
             time(&now);
-            command = (sb->input[0] == '/' &&
-                        ((sb->input[1] && sb->input[1] != '/') || (!sb->input[1] && c != '/'))) ||
-                      (!sb->input[0] && c == '/');
+            command = (sb->EB.text[0] == '/' &&
+                        ((sb->EB.text[1] && sb->EB.text[1] != '/') || (!sb->EB.text[1] && c != '/'))) ||
+                      (!sb->EB.text[0] && c == '/');
             if (!command && now - sb->tm_last_char > Config.time_user_types) {
                 sb->tm_last_char = now;
-                if (sb->thrid != -1 && sb->PL.count > 0)
+                if (sb->thrid != (pthread_t) -1 && sb->PL.count > 0)
                     msn_msg_typenotif(sb->sfd, sb->tid++, msn.login);
                 if (sb->sfd > -1) {
                     if (sb->PL.count == 0 && sb->master[0]) { /* call other party again */
@@ -278,11 +323,12 @@ void sb_type(int c)
                     Write(msn.nfd, s, strlen(s));
                 }
             }
-            eb_keydown(&sb->EB, c);
+            res = eb_keydown(&sb->EB, c);
             draw_sbebox(sb, 1);
     }
     UNLOCK(&SX);
     UNLOCK(&msn.lock);
+    return res;
 }
 
 void fcopy(FILE *s, FILE *d)
@@ -300,7 +346,7 @@ void sb_cleanup(void *arg)
     close(sb->sfd);
     sb->sfd = -1;
     msn_clist_free(&sb->PL);
-    sb->thrid = -1;
+    sb->thrid = (pthread_t) -1;
     sb->tid = 0;
     if (sb->fp_log != NULL) { /* copy temp log file */
         if (sb->master[0]) {
@@ -349,11 +395,11 @@ void msn_sb_handle_msgbody(char *arg1, char *arg2, char *msgdata, int len)
 
     sb->pending = 0;
     if (strchr(arg1, '@') != NULL) url2str(arg2, nick);
-    else sprintf(nick, "<%s>", getnick(arg1, arg1, Config.aliases));
+    else sprintf(nick, "<%s>", getnick1(arg1));
     contype[0] = 0;
     s = strafter(msgdata, "Content-Type: ");
     if (s == NULL) {
-        dlg(C_DBG, msgdata);
+        dlgn(C_DBG, len+2, "%s\n", msgdata);
         return;
     }
     sscanf(s, "%s", contype);
@@ -361,15 +407,26 @@ void msn_sb_handle_msgbody(char *arg1, char *arg2, char *msgdata, int len)
     if (strstr(contype, "text/plain") != NULL) {
         
     /* user chats */
-        char *text;
+        char *text, *e;
         
         s = strafter(msgdata, "\r\n\r\n");
         text = (char *) Malloc(len - (s - msgdata) + 1);
         sb->pending = 1;
         
         utf8decode(sb->ic, s, text);
-        dlg(C_DBG, "%s says:\n", getnick(arg1, nick, Config.aliases));
-        dlg(C_MSG, "%s\n", text);
+        if (says_ok(sb, arg1))
+            dlg(C_DBG, "%s says:\n", getnick2(arg1, nick));
+        dlgn(C_MSG, len+2, "%s\n", text);
+        if ((e = strstr(text, "http://")) != NULL) {
+            char url[SML];
+            int len = 7;
+            strncpy(url, e, 7); e += 7;
+            while (*e > 32 && len < SML) url[len++] = *(e++); /* clumsy parsing? :-) */
+            url[len] = 0;
+            xfl_add_url(&XL, msn.login, arg1, lrand48(), NULL, url);
+            draw_xfer(1);
+        }
+/*        playsound(SND_MSG); */
         free(text);
         
         return;
@@ -381,7 +438,7 @@ void msn_sb_handle_msgbody(char *arg1, char *arg2, char *msgdata, int len)
         time_t now;
         
         time(&now);
-        msn_clist_update(&msn.FL, arg1, NULL, -1, -1, now);
+        msn_clist_update(&msn.CL, msn_FL, arg1, NULL, -1, now, 0);
         draw_lst(1);
         /* rather tricky! (in order to be drawn *after* time_user_types sec) */
         sched_draw_lst(Config.time_user_types + 1); 
@@ -410,15 +467,16 @@ void msn_sb_handle_msgbody(char *arg1, char *arg2, char *msgdata, int len)
             else if (strcmp(cmd, "MSG") == 0) {
                 char *text = strdup(args);
                 utf8decode(msn_ic[0], args, text);
-                msg(C_DBG, "%s says:\n", getnick(arg1, nick, Config.aliases));
-                msg(C_MSG, "%s\n", text);
+                msg(C_DBG, "%s says:\n", getnick2(arg1, nick));
+                msgn(C_MSG, len+2, "%s\n", text);
                 free(text);
             } else if (strcmp(cmd, "DLG") == 0) {
                 char *text = strdup(args);
                 utf8decode(sb->ic, args, text);
                 sb->pending = 1;
-                dlg(C_DBG, "%s says:\n", getnick(arg1, nick, Config.aliases));
-                dlg(C_MSG, "%s\n", text);
+                dlg(C_DBG, "%s says:\n", getnick2(arg1, nick));
+                dlgn(C_MSG, len+2, "%s\n", text);
+/*                playsound(SND_MSG);   */
                 free(text);
             }
         }
@@ -461,10 +519,11 @@ void msn_sb_handle_msgbody(char *arg1, char *arg2, char *msgdata, int len)
                     xfl_add_file(&XL, XS_INVITE, msn.login, arg1, 1, cookie, sb, bfname, size);
                     UNLOCK(&XX);
                     sb->pending = 1;
-                    dlg(C_DBG, "%s wants to send you the file '%s', %u bytes long\n", getnick(arg1, nick, Config.aliases), fname, size);
+                    dlg(C_DBG, "%s wants to send you the file '%s', %u bytes long\n", getnick2(arg1, nick), fname, size);
+                    draw_xfer(1);
                 } else {
                 /* reject all other types of invitations */
-                    dlg(C_DBG, msgdata);
+                    dlgn(C_DBG, len+2, "%s\n", msgdata);
                     msn_msg_cancel(sb->sfd, sb->tid++, cookie, "REJECT_NOT_INSTALLED");
                 }
                 draw_sb(sb, 1);
@@ -495,14 +554,14 @@ void msn_sb_handle_msgbody(char *arg1, char *arg2, char *msgdata, int len)
 
                     x->data.file.port = port;
                     x->data.file.auth_cookie = auth;
-                    strcpy(x->data.file.ipaddr, ip);
+                    Strcpy(x->data.file.ipaddr, ip, SML);
                     x->status = xs;
                     pthread_create(&thr, NULL, msnftp_client, (void *) x);
                     pthread_detach(thr);
                 } else {
                     x->status = xs;
-                    strcpy(x->data.file.ipaddr, MyIP);
-                    strcpy(x->remote, arg1);
+                    Strcpy(x->data.file.ipaddr, MyIP, SML);
+                    Strcpy(x->remote, arg1, SML);
                     x->data.file.port = Config.msnftpd;
                     x->data.file.auth_cookie = lrand48();
                     msn_msg_accept2(sb->sfd, sb->tid++, x->inv_cookie, x->data.file.ipaddr, 
@@ -523,9 +582,9 @@ void msn_sb_handle_msgbody(char *arg1, char *arg2, char *msgdata, int len)
 
                 if (x->xclass == XF_FILE) {
                     if (x->incoming)
-                        dlg(C_DBG, "%s is not sending you '%s' any more\n", getnick(arg1, nick, Config.aliases), x->data.file.fname);
+                        dlg(C_DBG, "%s is not sending you '%s' any more\n", getnick2(arg1, nick), x->data.file.fname);
                     else
-                        dlg(C_DBG, "%s is not receiving '%s'\n", getnick(arg1, nick, Config.aliases), x->data.file.fname);
+                        dlg(C_DBG, "%s is not receiving '%s'\n", getnick2(arg1, nick), x->data.file.fname);
                 }
                 x->status = xs;
             } else {
@@ -533,13 +592,13 @@ void msn_sb_handle_msgbody(char *arg1, char *arg2, char *msgdata, int len)
             }
             UNLOCK(&XX);
         } else {
-            dlg(C_DBG, msgdata);
+            dlgn(C_DBG, len+2, "%s\n", msgdata);
         }
         draw_sb(sb, 1);
         return;
     }
     
-    if (Config.msg_debug == 2) dlg(C_DBG, msgdata);
+    if (Config.msg_debug == 2) dlgn(C_DBG, len+2, "%s\n", msgdata);
     else if (Config.msg_debug == 1) dlg(C_DBG, "MSG: %s\n", contype);
     
     if (Config.msg_debug > 0) sb->pending = 1;
@@ -560,6 +619,7 @@ void *msn_sbdaemon(void *arg)
     pthread_setspecific(key_sboard, arg);
     LOCK(&SX);
     tm = time(NULL);
+    s[0] = 0;
     ctime_r(&tm, s);
     dlg(C_NORMAL, "Switchboard session begin on %s", s);
     if (sb->called) dlg(C_DBG, "Invitation from <%s>\n", sb->master);
@@ -567,7 +627,7 @@ void *msn_sbdaemon(void *arg)
     sb->sfd = do_connect_sb(sb->sbaddr);
     UNLOCK(&SX);
     if (sb->sfd < 0) {
-        sb->thrid = -1;
+        sb->thrid = (pthread_t) -1;
         return NULL;
     }
     if (sb->called) {
@@ -586,9 +646,13 @@ void *msn_sbdaemon(void *arg)
     pthread_cleanup_push(sb_cleanup, arg);
     while (1) {
         r = bfParseLine(&buf, s, SXL);
+        pthread_testcancel();
         LOCK(&SX);
         if (r == -2) dlg(C_ERR, "bfParseLine(): buffer overrun\n");
-        else if (r < 0) dlg(C_ERR, "bfParseLine(): %s\n", strerror(errno));
+        else if (r < 0) {
+            if (errno != ECONNRESET || (errno == ECONNRESET && Config.err_connreset == 0))
+                dlg(C_ERR, "bfParseLine(): %s\n", strerror(errno));
+        }
         UNLOCK(&SX);
         if (r <= 0) break;
 
@@ -633,17 +697,17 @@ void *msn_sbdaemon(void *arg)
 
         /* user joins conversation */
 
-            sscanf(s + 4, "%s %s", arg1, arg2);
+            sscanf(s + 4, "%s %s", arg1, arg2); /* MSNP: capabilities ignored */
             url2str(arg2, nick);
             LOCK(&SX);
-            msn_clist_add(&sb->PL, -1, arg1, nick);
+            msn_clist_add(&sb->PL, 0, arg1, nick, NULL, NULL);
             if (sb->PL.count > 1) sb->multi = 1;
             if (sb->PL.count == 1) {
-                strcpy(sb->master, arg1);
-                strcpy(sb->mnick, nick);
+                Strcpy(sb->master, arg1, SML);
+                Strcpy(sb->mnick, nick, SML);
             }
             draw_plst(sb, 0);
-            dlg(C_DBG, "%s <%s> has joined the conversation\n", getnick(arg1, nick, Config.aliases), arg1);
+            dlg(C_DBG, "%s <%s> has joined the conversation\n", getnick2(arg1, nick), arg1);
             UNLOCK(&SX);
         } else if (is3(com, "IRO")) {
 
@@ -651,13 +715,13 @@ void *msn_sbdaemon(void *arg)
 
             int index;
 
-            sscanf(s + 4, "%*u %d %*d %s %s", &index, arg1, arg2);
+            sscanf(s + 4, "%*u %d %*d %s %s", &index, arg1, arg2); /* MSNP: capabilities ignored */
             url2str(arg2, nick);
             LOCK(&SX);
-            msn_clist_add(&sb->PL, -1, arg1, nick);
+            msn_clist_add(&sb->PL, 0, arg1, nick, NULL, NULL);
             draw_plst(sb, 0);
             if (sb->PL.count > 1) sb->multi = 1;
-            dlg(C_DBG, "%d. %s <%s>\n", index, getnick(arg1, nick, Config.aliases), arg1);
+            dlg(C_DBG, "%d. %s <%s>\n", index, getnick2(arg1, nick), arg1);
             UNLOCK(&SX);
         } else if (is3(com, "USR")) {
             LOCK(&SX);
@@ -678,7 +742,7 @@ void *msn_sbdaemon(void *arg)
 
             sscanf(s + 4, "%*u %*s %s", arg1);
             LOCK(&SX);
-            strcpy(sb->sessid, arg1);
+            Strcpy(sb->sessid, arg1, SML);
             dlg(C_DBG, "Calling other party ...\n");
             UNLOCK(&SX);
         } else if (is3(com, "BYE")) {
@@ -690,8 +754,8 @@ void *msn_sbdaemon(void *arg)
             sscanf(s + 4, "%s", arg1);
             LOCK(&msn.lock);
             LOCK(&SX);
-            msn_clist_rem(&sb->PL, arg1, -1);
-	    /* gtmess is a GreaT MESS of source code! ;-) [egg] */
+            msn_clist_rem(&sb->PL, 0, arg1, NULL);
+	    /* GreaT MESS of source code! ;-) */
             leave = sb->PL.count == 0 && sb->multi;
             draw_plst(sb, 0);
             dlg(C_ERR, "<%s> has left the conversation\n", arg1);
@@ -714,13 +778,14 @@ void *msn_sbdaemon(void *arg)
         /* error code */
 
             LOCK(&SX);
-            sb->pending = 1;
+            if (SL.head != sb) sb->pending = 1;
             if (!scan_error_sb(com))
-                dlg(C_DBG, "<<< %s", s);
+                dlgn(C_DBG, strlen(s)+6, "<<< %s\n", s);
             UNLOCK(&SX);
         }
     }
     tm = time(NULL);
+    s[0] = 0;
     ctime_r(&tm, s);
     dlg(C_ERR, "Disconnected from switchboard on %s", s);
 /*    sb->pending = 1;*/
@@ -743,22 +808,22 @@ void sboard_openlog(msn_sboard_t *p)
     else p->fp_log = NULL;
 }
 
-msn_sboard_t *msn_sblist_add(msn_sblist_t *q, char *sbaddr, char *hash,
-                             char *master, char *mnick, int called, char *sessid)
+msn_sboard_t *msn_sblist_add(msn_sblist_t *q, const char *sbaddr, const char *hash,
+                             const char *master, const char *mnick, int called, const char *sessid)
 {
     msn_sboard_t *p;
 
     p = (msn_sboard_t *) Malloc(sizeof(msn_sboard_t));
-    strcpy(p->sbaddr, sbaddr);
-    strcpy(p->hash, hash);
+    Strcpy(p->sbaddr, sbaddr, SML);
+    Strcpy(p->hash, hash, HDATASZ);
     p->called = called;
     sboard_openlog(p);
 
     p->sessid[0] = 0;
-    if (sessid != NULL) strcpy(p->sessid, sessid);
+    if (sessid != NULL) Strcpy(p->sessid, sessid, SML);
     p->PL.head = NULL; p->PL.count = 0;
-    strcpy(p->master, master);
-    strcpy(p->mnick, mnick);
+    Strcpy(p->master, master, SML);
+    Strcpy(p->mnick, mnick, SML);
     p->multi = 0;
 
     p->w_dlg.w = W_DLG_W;
@@ -783,16 +848,20 @@ msn_sboard_t *msn_sblist_add(msn_sblist_t *q, char *sbaddr, char *hash,
     else p->ic = (iconv_t) -1;
     p->tid = 0;
     p->sfd = -1;
-    p->thrid = -1;
-    p->input[0] = 0;
-    eb_init(&p->EB, SXL - 1, SXL - 1, w_msg.w, p->input, p->winput);
+    p->thrid = (pthread_t) -1;
+    eb_init(&p->EB, SXL-1, w_msg.w);
+    p->EB.grow = 1;
     p->EB.utf8 = utf8_mode;
     p->EB.history = 1;
+    eb_history_add(&p->EB, ZS, 0);
 /*    eb_settext(&p->EB, ZS);*/
     p->pending = 0;
     p->tm_last_char = 0;
     p->plskip = 0;
     p->dlg_now = 0;
+    p->real_dlg_now = 0;
+    p->last_nick[0] = 0;
+    p->last_spoken = 0;
 
     if (q->head == NULL) {
         /* empty list */
@@ -817,10 +886,11 @@ int msn_sblist_rem(msn_sblist_t *q)
     xfer_t *x;
 
     if (p == NULL) return 0;
-    if (p->thrid != -1) {
+    if (p->thrid != (pthread_t) -1) {
 /*        close(p->sfd);*/
-        pthread_cancel(p->thrid);
-        pthread_join(p->thrid, NULL);
+        pthread_t tmp = p->thrid;
+        pthread_cancel(tmp);
+        pthread_join(tmp, NULL);
     }
     msn_clist_free(&p->PL);
     delwin(p->w_dlg.wh);
@@ -835,7 +905,7 @@ int msn_sblist_rem(msn_sblist_t *q)
     if (q->start == p) q->start = NULL;
     q->head = p->next;
     if (q->head == p) q->head = NULL; /* only one element */
-    hlist_free(&p->EB.HL);
+    eb_free(&p->EB);
     LOCK(&XX);
        for (x = XL.head; x != NULL; x = x->next) if (x->sb == p) {
            x->sb = NULL;
