@@ -2,7 +2,7 @@
  *    screen.c
  *
  *    gtmess - MSN Messenger client
- *    Copyright (C) 2002-2005  George M. Tzoumas
+ *    Copyright (C) 2002-2011  George M. Tzoumas
  *
  *    This program is free software; you can redistribute it and/or modify
  *    it under the terms of the GNU General Public License as published by
@@ -30,29 +30,50 @@
 #include"unotif.h"
 #include"xfer.h"
 #include"utf8.h"
+#include"queue.h"
 #include"gtmess.h"
 
-int attrs[] = {A_NORMAL, A_UNDERLINE, A_BOLD, A_NORMAL, A_STANDOUT, A_REVERSE, A_NORMAL};
+int C_NORMAL = A_NORMAL;
+int C_ERR    = A_UNDERLINE;
+int C_DBG    = A_BOLD;
+int C_MSG    = A_NORMAL;
+int C_MNU    = A_REVERSE;
+int C_EBX    = A_NORMAL;
+int C_GRP    = A_UNDERLINE;
+
 int statattrs[] = {A_REVERSE, A_REVERSE, A_STANDOUT, A_REVERSE,
         A_REVERSE, A_REVERSE, A_REVERSE, A_REVERSE, A_REVERSE, A_REVERSE};
-int lstatattrs[] = {A_NORMAL, A_NORMAL, A_REVERSE, A_BOLD, A_BOLD,
-        A_BOLD | A_UNDERLINE, A_BOLD, A_BOLD | A_UNDERLINE, A_BOLD, A_STANDOUT};
+int lstatattrs[] = {A_NORMAL, A_NORMAL, A_BOLD, A_BOLD, A_BOLD,
+        A_BOLD, A_BOLD, A_BOLD, A_BOLD, A_UNDERLINE};
 
 TWindow w_msg, w_lst, w_back, w_xfer;
 pthread_mutex_t scr_lock;
+int scr_shutdown = 0;
 
 int wvis; /* 0 = sb window visible, 1 = transfers window visible */
+int clmenu; /* CL menu active */
 int w_msg_top = 0;
 
 /* last arguments to msg2() */
 char msg2_str[SML] = {0};
-cattr_t msg2_attr = C_MNU;
+int msg2_attr;
 
 time_t msg_now = 0;
-extern pthread_mutex_t time_lock;
+
+extern struct tm now_tm;
+
+
+/* text message queue before screen init. */
+xqueue_t msg_q; 
 
 int fullscreen = 0;
 
+float scr_rx = 0.75;
+float scr_ry = 0.4;
+
+int SCOLS, SLINES;
+int utf8_mode;
+
 char *get_next_word(char **text, char **word, int *len, int *numlines)
 {
     char *s = *text;
@@ -82,7 +103,8 @@ void wprintx(WINDOW *w, char *txt, int indent)
 {
     int maxwidth, ymax, width, wlen, nlines;
     char *text, *word;
-    char tw[SXL];
+    char twbuf[SXL], *tw;
+    int alloc;
     
     getmaxyx(w, ymax, maxwidth);
     maxwidth--;
@@ -90,51 +112,53 @@ void wprintx(WINDOW *w, char *txt, int indent)
     width = indent;
     while (get_next_word(&text, &word, &wlen, &nlines) != NULL) {
         if (width > 0 && width+wlen >= maxwidth) waddch(w, '\n');
+        tw = twbuf; alloc = 0;
+        if (wlen >= SXL) {
+            tw = (char *) malloc(wlen+1);
+            alloc = 1;
+        }
         strncpy(tw, word, wlen);
         tw[wlen] = 0;
         if (nlines > 0) for (; nlines > 0; --nlines) waddch(w, '\n');
         waddstr(w, tw);
         waddch(w, ' ');
         getyx(w, ymax, width);
+        if (alloc && tw != NULL) free(tw);
     }
     for (; nlines > 0; --nlines) waddch(w, '\n');
 }
 
-void gs_draw_prompt(cattr_t attr, const char *prompt)
+void gs_draw_prompt(int attr, const char *prompt)
 {
     move(SLINES - 1, 0);
-    bkgdset(' ' | attrs[attr]);
+    bkgdset(' ' | attr);
     clrtoeol();
-    bkgdset(' ' | attrs[C_NORMAL]);
+    bkgdset(' ' | C_NORMAL);
     move(SLINES - 1, 0);
-    attrset(attrs[attr] ^ A_REVERSE);
+    attrset(attr ^ A_REVERSE);
     addstr(prompt);
-    attrset(attrs[C_NORMAL]);
+    attrset(C_NORMAL);
 }
 
-void gs_draw_ebox(ebox_t *e, cattr_t attr, int pl)
+void gs_draw_ebox(ebox_t *e, int attr, int pl)
 {
     move(SLINES - 1, pl);
-    attrset(attrs[attr]);
+    attrset(attr);
     eb_draw(e, stdscr);
-    attrset(attrs[C_NORMAL]);
+    attrset(C_NORMAL);
 }
 
-int get_string(cattr_t attr, int mask, const char *prompt, char *dest)
+int get_string(int attr, int mask, const char *prompt, char *dest, size_t n)
 {
     int r;
-    char s[SML];
-    wchar_t ws[SML];
     int c, pl;
     ebox_t E;
 
-    strncpy(s, dest, SML);
-    s[SML - 1] = 0;
     pl = strlen(prompt);
-    eb_init(&E, SML-1, SML-1, SCOLS-1-pl, s, ws);
+    eb_init(&E, SML-1, SCOLS-1-pl);
     E.utf8 = utf8_mode;
     E.mask = mask;
-    eb_settext(&E, s);
+    eb_settext(&E, dest);
     c = -1;
     
     gs_draw_prompt(attr, prompt);    
@@ -153,43 +177,55 @@ int get_string(cattr_t attr, int mask, const char *prompt, char *dest)
     }
     r = (c == 13);
     if (r) {
-        if (strcmp(dest, s) == 0) r <<= 1;
-        strcpy(dest, s);
+        if (strcmp(dest, (char *) E.text) == 0) r <<= 1;
+        else Strcpy(dest, (char *) E.text, n);
     }
     move(SLINES - 1, SCOLS - 1);
+    eb_free(&E);
     return r;
 }
 
 /* print a message */
-void vwmsg(TWindow *w, time_t *sbtime, cattr_t attr, FILE *fp_log, const char *fmt, va_list ap)
+int vwmsg(TWindow *w, int size, time_t *sbtime, time_t *real_sbtime, int attr, 
+        FILE *fp_log, const char *fmt, va_list ap)
 {
-    char tmp[SXL], *s;
+    char tmpbuf[SXL], *tmp, *s;
+    int alloc;
     time_t now2;
     int indent = 0;
+    int trunc;
 
-    vsprintf(tmp, fmt, ap);
+    if (size < SXL) {
+        tmp = tmpbuf;
+        alloc = 0;
+    } else {
+        tmp = (char *) malloc(size);
+        alloc = 1;
+    }
+
+    trunc = vsnprintf(tmp, size, fmt, ap);
+    if (trunc < size) trunc = 0; else tmp[size-1] = 0;
 
     for (s = tmp; *s; s++) if (*s == '\r') *s = ' ';
 
     LOCK(&w->lock);
-    wattrset(w->wh, attrs[attr]);
+    wattrset(w->wh, attr);
     if (fp_log != NULL) {
-        switch (attr) {
-            case C_DBG: 
-                fprintf(fp_log, "(--) "); break;
-            case C_ERR: 
-                fprintf(fp_log, "(!!) "); break;
-            default:
-                fprintf(fp_log, "     "); break;
-        }
+        if (attr == C_DBG)
+            fprintf(fp_log, "(--) "); 
+        else if (attr == C_ERR)
+            fprintf(fp_log, "(!!) ");
+        else fprintf(fp_log, "     ");
         fflush(fp_log);
     }
 /*    waddstr(w->wh, tmp);*/
     time(&now2);
+    /* BUG: ugly hack to ignore "disconnected..." msgs */
+    if (attr != C_ERR) *real_sbtime = now2;
     LOCK(&time_lock);
     if (difftime(now2, *sbtime) > 60) {
         *sbtime = now2;
-        sbtime -= now2 % 60;
+        *sbtime -= now2 % 60;
         wprintw(w->wh, "[%02d:%02d] ", now_tm.tm_hour, now_tm.tm_min);
         if (fp_log != NULL) {
             fprintf(fp_log, "[%02d:%02d] ", now_tm.tm_hour, now_tm.tm_min);
@@ -199,45 +235,90 @@ void vwmsg(TWindow *w, time_t *sbtime, cattr_t attr, FILE *fp_log, const char *f
     } else indent = 0;
     UNLOCK(&time_lock);
     wprintx(w->wh, tmp, indent);
-    wattrset(w->wh, attrs[C_NORMAL]);
+    wattrset(w->wh, C_NORMAL);
     UNLOCK(&w->lock);
     if (fp_log != NULL) {
         fputs(tmp, fp_log);
         fflush(fp_log);
     }
+    
+    if (alloc && tmp != NULL) free(tmp);
+    return trunc;
 }
 
 /* message window */
-void msg(cattr_t attr, const char *fmt, ...)
+void msg(int attr, const char *fmt, ...)
 {
     va_list ap;
-    char txt[SML];
+    int r;
+    
+    va_start(ap, fmt);
+    r = vmsg(attr, SML, fmt, ap);
+    va_end(ap);
+    /* workaround for empty va_list for ARM --
+         infinite recursion should not happen anyway :) */
+    if (r) msg(C_ERR, "msg(): output truncated\n");
+}
+
+void msgn(int attr, int size, const char *fmt, ...)
+{
+    va_list ap;
+    int r;
+    
+    va_start(ap, fmt);
+    r = vmsg(attr, size, fmt, ap);
+    va_end(ap);
+    if (r) msg(C_ERR, "msgn(): output truncated\n");
+}
+
+int screen_shut()
+{
+    int v = 0;
+    LOCK(&scr_lock);
+    v = scr_shutdown;
+    UNLOCK(&scr_lock);
+    return v;
+}
+
+int vmsg(int attr, int size, const char *fmt, va_list ap)
+{
+    char txtbuf[SML], *txt;
+    int alloc;
     char *s;
     int y, x, newtop;
     time_t now2;
     int indent = 0;
+    int trunc = 0;
     
-    va_start(ap, fmt);
-    vsprintf(txt, fmt, ap);
-    va_end(ap);
+    if (screen_shut()) return 0;
+    
+    if (size < SML) {
+        txt = txtbuf;
+        alloc = 0;
+    } else {
+        txt = (char *) malloc(size);
+        alloc = 1;
+    }
+    trunc = vsnprintf(txt, size, fmt, ap);
+    if (trunc < size) trunc = 0; else txt[size-1] = 0;
+
     for (s = txt; *s; s++) if (*s == '\r') *s = ' ';
     
     if (!fullscreen) {
+        queue_put(&msg_q, 0, txt, strlen(txt) + 1);
         fprintf(stderr, "%s\n", txt);
-        return;
+        if (alloc && txt != NULL) free(txt);
+        return 0;
     }
     
     LOCK(&w_msg.lock);
-    wattrset(w_msg.wh, attrs[attr]);
+    wattrset(w_msg.wh, attr);
     if (msn.fp_log != NULL) {
-        switch (attr) {
-            case C_DBG: 
-                fprintf(msn.fp_log, "(--) "); break;
-            case C_ERR: 
-                fprintf(msn.fp_log, "(!!) "); break;
-            default:
-                fprintf(msn.fp_log, "     "); break;
-        }
+        if (attr == C_DBG)
+            fprintf(msn.fp_log, "(--) ");
+        else if (attr == C_ERR)
+            fprintf(msn.fp_log, "(!!) ");
+        else fprintf(msn.fp_log, "     ");
         fflush(msn.fp_log);
     }
     time(&now2);
@@ -258,42 +339,50 @@ void msg(cattr_t attr, const char *fmt, ...)
         fputs(txt, msn.fp_log);
         fflush(msn.fp_log);
     }
-    wattrset(w_msg.wh, attrs[C_NORMAL]);
+    wattrset(w_msg.wh, C_NORMAL);
     getyx(w_msg.wh, y, x);
     newtop = y - w_msg.h + 1;
     if (w_msg_top < newtop) w_msg_top = newtop;
     if (w_msg_top < 0) w_msg_top = 0;
+    LOCK(&scr_lock);
     getyx(stdscr, y, x);
     copywin(w_msg.wh, stdscr, w_msg_top, 0, w_msg.y, w_msg.x,
             w_msg.y + w_msg.h - 1, w_msg.x + w_msg.w - 1, FALSE);
     UNLOCK(&w_msg.lock);
     move(y, x);
     refresh();
+    UNLOCK(&scr_lock);
+    if (alloc && txt != NULL) free(txt);
+    return trunc;
 }
 
-void msg2(cattr_t attr, const char *fmt, ...)
+void msg2(int attr, const char *fmt, ...)
 {
     va_list ap;
     char txt[SML];
     int y, x;
 
+    if (screen_shut()) return;
+
     va_start(ap, fmt);
     vsprintf(txt, fmt, ap);
     va_end(ap);
-    strcpy(msg2_str, txt);
+    Strcpy(msg2_str, txt, SML);
     msg2_attr = attr;
     if (utf8_mode) txt[widthoffset(txt, SCOLS-2)] = 0;
     else txt[SCOLS-2] = 0;
+    LOCK(&scr_lock);
     getyx(stdscr, y, x);
     move(SLINES - 1, 0);
-    bkgdset(' ' | attrs[attr]);
+    bkgdset(' ' | attr);
     clrtoeol();
-    bkgdset(' ' | attrs[C_NORMAL]);
-    attrset(attrs[attr]);
+    bkgdset(' ' | C_NORMAL);
+    attrset(attr);
     addstr(txt);
-    attrset(attrs[C_NORMAL]);
+    attrset(C_NORMAL);
     move(y, x);
     refresh();
+    UNLOCK(&scr_lock);
 }
 
 void draw_rest()
@@ -301,9 +390,9 @@ void draw_rest()
     int y, x;
 
     getyx(stdscr, y, x);
-    move(1, SCOLS * 3 / 4);
+    move(1, (int) (SCOLS * scr_rx + 0.5));
     vline(ACS_VLINE, SLINES - 2);
-    mvaddch(w_msg.y - 1, SCOLS * 3 / 4, ACS_RTEE);
+    mvaddch(w_msg.y - 1, (int) (SCOLS * scr_rx + 0.5), ACS_RTEE);
     move(y, x);
 }
 
@@ -312,7 +401,10 @@ void draw_status(int r)
     int y, x;
     char txt[SML];
     
-    sprintf(txt, "%s (%s)", getnick(msn.login, msn.nick, Config.aliases), msn_stat_name[msn.status]);
+    if (msn.psm[0])
+        sprintf(txt, "%s (%s) - %s", msn.nick, msn_stat_name[msn.status], msn.psm);
+    else
+        sprintf(txt, "%s (%s)", msn.nick, msn_stat_name[msn.status]);
     if (utf8_mode) txt[widthoffset(txt, SCOLS-6)] = 0;
     else txt[SCOLS-6] = 0;
     getyx(stdscr, y, x);
@@ -321,13 +413,13 @@ void draw_status(int r)
     printw("%*c", SCOLS-5, ' ');
 /*    bkgdset(' ' | statattrs[msn.status]);
     clrtoeol();
-    bkgdset(' ' | attrs[C_NORMAL]);*/
+    bkgdset(' ' | C_NORMAL);*/
     move(0, 0);
     addstr(txt); 
 /*    addch(' ');
     addch(ACS_VLINE); 
     printw(" gtmess %s", VERSION);*/
-    attrset(attrs[C_NORMAL]);
+    attrset(C_NORMAL);
     move(y, x);
     draw_time(r);
 }
@@ -340,17 +432,18 @@ void draw_time(int r)
     move(0, SCOLS-5);
     bkgdset(' ' | statattrs[msn.status]);
     clrtoeol();
-    bkgdset(' ' | attrs[C_NORMAL]);
+    bkgdset(' ' | C_NORMAL);
     attrset(statattrs[msn.status]);
     LOCK(&time_lock);
     printw("%02d:%02d", now_tm.tm_hour, now_tm.tm_min);
     UNLOCK(&time_lock);
-    attrset(attrs[C_NORMAL]);
+    attrset(C_NORMAL);
     move(y, x);
     if (r) refresh();
 }
 
 
+/* VERY tricky drawing routine, since data are not static :-/ */
 void draw_lst(int r)
 {
     msn_contact_t *p;
@@ -358,38 +451,145 @@ void draw_lst(int r)
     time_t now;
     char ch;
     int lines, total;
-    int y, x;
-
-    LOCK(&w_lst.lock);
-    wclear(w_lst.wh);
-
-    lines = 0;
-    total = msn.GL.count;
-    if (Config.online_only == 0) total += msn.FL.count;
-    else 
-        for (p = msn.FL.head; p != NULL; p = p->next)
-            if (p->status >= MS_NLN) total++;
-    time(&now);
-    for (g = msn.GL.head; g != NULL; g = g->next) {
-        if (lines >= msn.flskip && lines < total + msn.flskip) {
-            wattrset(w_lst.wh, attrs[C_GRP]);
-            wprintw(w_lst.wh, "%s\n", g->name);
-        }
-        lines++;
-        for (p = msn.FL.head; p != NULL; p = p->next)
-            if (p->gid == g->gid) {
-                if (Config.online_only && p->status < MS_NLN) continue;
-                if (lines >= msn.flskip && lines < total + msn.flskip) {
-                    wattrset(w_lst.wh, lstatattrs[p->status]);
-                    if (now - p->tm_last_char <= Config.time_user_types) ch = '!';
-                    else if (p->blocked) ch = '+';
-                    else ch = ' ';
-                    wprintw(w_lst.wh, "%c%c %s\n", ch, msn_stat_char[p->status], 
-                            getnick(p->login,p->nick, Config.aliases));
-                }
-                lines++;
+    int y, x, i, j, ingroup;
+    int cur_grp, cur_con;
+    int gcount = msn.GL.count + 1;
+    msn_group_t *b[gcount];
+    msn_contact_t **a[gcount];
+    int gsize[gcount];
+    int found;
+    
+    total = gcount;
+    cur_grp = 0;
+    cur_con = 0;
+    found = 0;
+    if (Config.online_only == 0) {
+        for (p = msn.CL.head; p != NULL; p = p->next)
+            if (p->lflags & msn_FL) {
+                if (p->groups > 0) total += p->groups;
+                else total++;
+            }
+    } else {
+        for (p = msn.CL.head; p != NULL; p = p->next)
+            if ((p->lflags & msn_FL) && p->status >= MS_NLN) {
+                if (p->groups > 0) total += p->groups;
+                else total++;
             }
     }
+    
+    /* create arrays for random access */
+    for (i = 0, g = msn.GL.head; g != NULL; g = g->next, i++) {
+        b[i] = g;
+        gsize[i] = 0;
+        for (p = msn.CL.head; p != NULL; p = p->next) { /* first pass to compute size :-/ */
+            if ((p->lflags & msn_FL) == 0 || (Config.online_only && p->status < MS_NLN)) continue;
+            for (j = 0; j < p->groups; j++) {
+                gsize[i] += strcmp(p->gid[j], g->gid) == 0;
+            }
+        }
+        if (gsize[i]) {
+            /* is there an easy way to avoid malloc() all the time? perhaps a custom malloc :-) */
+            a[i] = (msn_contact_t **) malloc(gsize[i] * sizeof(msn_contact_t *));
+            ingroup = 0;
+            for (p = msn.CL.head; p != NULL; p = p->next) {
+                if ((p->lflags & msn_FL) == 0 || (Config.online_only && p->status < MS_NLN)) continue;
+                for (j = 0; j < p->groups; j++) if (strcmp(p->gid[j], g->gid) == 0) {
+                    a[i][ingroup] = p;
+                    if (strcmp(msn.hgid, g->gid) == 0 && strcmp(p->login, msn.hlogin) == 0)
+                        cur_grp = i, cur_con = ingroup, found = 1;
+                    ingroup++;
+                }
+            }
+        }
+    }
+    /* groupless contacts */
+    b[gcount-1] = &msn_group_0;
+    gsize[gcount-1] = 0;
+    for (p = msn.CL.head; p != NULL; p = p->next)
+        if ((p->lflags & msn_FL) && (!Config.online_only || p->status >= MS_NLN) && p->groups == 0)
+            gsize[gcount-1]++;
+    if (gsize[gcount-1]) {
+        a[gcount-1] = (msn_contact_t **) malloc(gsize[gcount-1] * sizeof(msn_contact_t *));
+        ingroup = 0;
+        for (p = msn.CL.head; p != NULL; p = p->next)
+            if ((p->lflags & msn_FL)
+                && (!Config.online_only || p->status >= MS_NLN) && p->groups == 0) {
+                a[gcount-1][ingroup] = p;
+                if (msn.hgid[0] == 0 && strcmp(p->login, msn.hlogin) == 0)
+                    cur_grp = gcount-1, cur_con = ingroup, found = 1;
+                ingroup++;
+            }
+    }
+
+    lines = 0;
+    
+    /* handle selection */
+    if (found) {
+        if (msn.dhid > 0) {
+            cur_con++;
+            while (1) {
+                if (cur_con >= gsize[cur_grp]) {
+                    cur_grp++;
+                    if (cur_grp >= gcount) cur_grp = 0;
+                    cur_con = 0; 
+                } else break;
+            }
+            msn.dhid = 0;
+        } else if (msn.dhid < 0) {
+            cur_con--;
+            while (1) {
+                if (cur_con < 0) {
+                    cur_grp--; 
+                    if (cur_grp < 0) cur_grp = gcount - 1;
+                    cur_con = gsize[cur_grp] - 1;
+                } else break;
+            }
+            msn.dhid = 0;
+        }
+    } else while (cur_grp < gcount) {
+        if (cur_con < gsize[cur_grp]) break;
+        else cur_grp++, cur_con = 0;
+    }
+
+    if (cur_grp >= gcount) {
+        clmenu = 0;
+    }
+    
+    if (clmenu) {
+        if (b[cur_grp]->gid[0]) strcpy(msn.hgid, b[cur_grp]->gid);
+        else msn.hgid[0] = 0;
+        strcpy(msn.hlogin, a[cur_grp][cur_con]->login);
+    }
+
+    /* do the drawing */
+    LOCK(&w_lst.lock);
+    wclear(w_lst.wh);
+    
+    time(&now);
+    for (i = 0; i < gcount; i++) {
+        if (lines >= msn.flskip && lines < total + msn.flskip) {
+            wattrset(w_lst.wh, C_GRP);
+            wprintw(w_lst.wh, "%s\n", b[i]->name);
+        }
+        lines++;
+        for (j = 0; j < gsize[i]; j++) {
+            if (lines >= msn.flskip && lines < total + msn.flskip) {
+                p = a[i][j];
+                if (clmenu && i == cur_grp && j == cur_con)
+                    wattrset(w_lst.wh, statattrs[p->status]);
+                else wattrset(w_lst.wh, lstatattrs[p->status]);
+                if (now - p->tm_last_char <= Config.time_user_types) ch = '!';
+                else if ((p->lflags & msn_BL)) ch = '+'; /* blocked */
+                else if (p->ignored) ch = ':';
+                else ch = ' ';
+                wprintw(w_lst.wh, "%c%c %s\n", ch, msn_stat_char[p->status], getnick1c(p));
+            }
+            lines++;
+        }
+    }
+
+    for (i = 0; i < gcount; i++) if (gsize[i]) free(a[i]);
+
     getyx(stdscr, y, x);
     copywin(w_lst.wh, stdscr, 0, 0, w_lst.y, w_lst.x,
             w_lst.y + w_lst.h - 1, w_lst.x + w_lst.w - 1, FALSE);
@@ -441,7 +641,7 @@ void draw_dlg(msn_sboard_t *sb, int r)
 void draw_plst(msn_sboard_t *sb, int r)
 {
     msn_contact_t *p;
-    TWindow *w, data;
+    TWindow *w;
     int lines, total;
     int y, x;
 
@@ -454,14 +654,14 @@ void draw_plst(msn_sboard_t *sb, int r)
         lines = 0;
         total = sb->PL.count + 1;
         if (lines >= sb->plskip && lines < total + sb->plskip) {
-            wattrset(w->wh, attrs[C_GRP]);
-            wprintw(w->wh, "[%s]\n\n", getnick(sb->master, sb->mnick, Config.aliases));
+            wattrset(w->wh, C_GRP);
+            wprintw(w->wh, "[%s]\n\n", getnick2(sb->master, sb->mnick));
         }
         lines++;
         for (p = sb->PL.head; p != NULL; p = p->next) {
             if (lines >= sb->plskip && lines < total + sb->plskip) {
                 wattrset(w->wh, lstatattrs[MS_UNK]);
-                wprintw(w->wh, "  %s\n", getnick(p->login, p->nick, Config.aliases));
+                wprintw(w->wh, "  %s\n", getnick1c(p));
             }
             lines++;
         }
@@ -489,7 +689,7 @@ void draw_sbbar(int r)
     int y, x, w;
     
     ch = &txt[0];
-    attr = attrs[C_NORMAL];
+    attr = C_NORMAL;
     
     s = SL.start;
     c = SL.count;
@@ -508,7 +708,7 @@ void draw_sbbar(int r)
             else if (s->pending) strcat(txt, "}-");
             else strcat(txt, " -");
         }
-        if (strlen(txt) > w_msg.w) {
+        if (strwidth(txt) > w_msg.w) {
             for (i = 0; i < c; i++, s = s->next) {
                if (s == SL.head) *ch++ = 'O';
                else if (s->pending) *ch++ = '+';
@@ -535,17 +735,38 @@ void draw_sbbar(int r)
     if (r) refresh();
 }
 
+void dlg(int attr, const char *fmt, ...)
+{
+    va_list ap;
+    int r;
+    
+    va_start(ap, fmt);
+    r = vdlg(attr, SXL, fmt, ap);
+    va_end(ap);
+    if (r) msg(C_ERR, "dlgn(): output truncated\n");
+}
+
+void dlgn(int attr, int size, const char *fmt, ...)
+{
+    va_list ap;
+    int r;
+    
+    va_start(ap, fmt);
+    r = vdlg(attr, size, fmt, ap);
+    va_end(ap);
+    if (r) msg(C_ERR, "dlgn(): output truncated\n");
+}
+
+
 /* switchboard conversation window */
-void dlg(cattr_t attr, const char *fmt, ...)
+int vdlg(int attr, int size, const char *fmt, va_list ap)
 {
     msn_sboard_t *sb = (msn_sboard_t *) pthread_getspecific(key_sboard);
     int y, x, newtop;
-    va_list ap;
+    int r;
 
-    if (sb == NULL) return;
-    va_start(ap, fmt);
-    vwmsg(&sb->w_dlg, &sb->dlg_now, attr, sb->fp_log, fmt, ap);
-    va_end(ap);
+    if (sb == NULL) return 0;
+    r = vwmsg(&sb->w_dlg, size, &sb->dlg_now, &sb->real_dlg_now, attr, sb->fp_log, fmt, ap);
     
     getyx(sb->w_dlg.wh, y, x);
     newtop = y - sb->w_dlg.h + 1;
@@ -554,6 +775,7 @@ void dlg(cattr_t attr, const char *fmt, ...)
 
     draw_dlg(sb, 0);
     draw_sbbar(1);
+    return r;
 }
 
 void draw_sbebox(msn_sboard_t *sb, int r)
@@ -563,14 +785,15 @@ void draw_sbebox(msn_sboard_t *sb, int r)
     
     move(w_msg.y - 2, 0);
     if (sb != NULL) {
-        attrset(attrs[C_EBX]);
+        attrset(C_EBX);
         eb_draw(&sb->EB, stdscr);
     } else {
-        attrset(attrs[C_NORMAL]);
+        attrset(C_NORMAL);
         printw("%*c", w_msg.w, ' ');
         move(SLINES - 1, SCOLS - 1);
     }
-    attrset(attrs[C_NORMAL]);
+    /* if (sb == NULL || clmenu == 1)  ... */
+    attrset(C_NORMAL);
     
     if (r) refresh();
 }
@@ -594,6 +817,25 @@ void draw_all()
     draw_rest(1);
 }
 
+void color_test()
+{
+    int i;
+    msg(C_NORMAL, "Normal text\n");
+    msg(C_ERR, "Error message\n");
+    msg(C_DBG, "Debugging message\n");
+    msg(C_MSG, "Server message\n");
+    msg(C_MNU, "Menu entry\n");
+    msg(C_EBX, "Editbox\n");
+    msg(C_GRP, "Group name\n");
+    
+    msg(C_NORMAL, "\nContact colors:\n");
+    for (i = MS_FLN; i <= MS_UNK; i++)
+        msg(lstatattrs[i], "%s\n", msn_stat_name[i]);
+    msg(C_NORMAL, "\nStatus colors:\n");
+    for (i = MS_FLN; i <= MS_UNK; i++)
+        msg(statattrs[i], "%s\n", msn_stat_name[i]);
+}
+
 void screen_resize()
 {
 #ifdef HAVE_WRESIZE
@@ -603,8 +845,8 @@ void screen_resize()
     if (SLINES < 9) SLINES = 9; else SLINES = LINES;
     
     LOCK(&w_msg.lock);
-    w_msg.w = SCOLS * 3 / 4;
-    w_msg.h = 2 * SLINES / 5 - 2;
+    w_msg.w = (int) (SCOLS * scr_rx + 0.5);
+    w_msg.h = (int) (SLINES * scr_ry + 0.5) - 2;
     wresize(w_msg.wh, W_MSG_LBUF, w_msg.w);
     w_msg.x = 0;
     w_msg.y = SLINES - 1 - w_msg.h;
@@ -658,27 +900,35 @@ void screen_resize()
 
 void screen_init(int colors)
 {
+    qelem_t *q;
+
     initscr();
     if (colors == 1 || (colors == 0 && has_colors())) {
         start_color();
-        init_pair(1, COLOR_RED, COLOR_BLACK);
+#ifdef HAVE_USE_DEFAULT_COLORS
+        use_default_colors();
+#define COLOR_BG -1
+#else
+#define COLOR_BG COLOR_BLACK
+#endif
+        init_pair(1, COLOR_RED, COLOR_BG);
         lstatattrs[MS_FLN] = lstatattrs[MS_HDN] =
-                attrs[C_ERR] = COLOR_PAIR(1);
+                C_ERR = COLOR_PAIR(1);
 
-        init_pair(2, COLOR_GREEN, COLOR_BLACK);
-        lstatattrs[MS_NLN] = attrs[C_DBG] = COLOR_PAIR(2);
+        init_pair(2, COLOR_GREEN, COLOR_BG);
+        lstatattrs[MS_NLN] = C_DBG = COLOR_PAIR(2);
 
-        init_pair(3, COLOR_CYAN, COLOR_BLACK);
-        lstatattrs[MS_BRB] = attrs[C_MSG] = COLOR_PAIR(3);
+        init_pair(3, COLOR_CYAN, COLOR_BG);
+        lstatattrs[MS_BRB] = C_MSG = COLOR_PAIR(3);
 
         init_pair(4, COLOR_WHITE, COLOR_RED);
         statattrs[MS_FLN] = COLOR_PAIR(4);
 
         init_pair(5, COLOR_BLUE, COLOR_CYAN);
-        statattrs[MS_NLN] = attrs[C_MNU] = COLOR_PAIR(5);
-        statattrs[MS_HDN] = attrs[C_EBX] = COLOR_PAIR(5) | A_REVERSE;
+        statattrs[MS_NLN] = C_MNU = COLOR_PAIR(5);
+        statattrs[MS_HDN] = C_EBX = COLOR_PAIR(5) | A_REVERSE;
 
-        init_pair(6, COLOR_MAGENTA, COLOR_BLACK);
+        init_pair(6, COLOR_MAGENTA, COLOR_BG);
         lstatattrs[MS_BSY] = lstatattrs[MS_PHN] = COLOR_PAIR(6);
 
         init_pair(7, COLOR_BLACK, COLOR_CYAN);
@@ -691,16 +941,18 @@ void screen_init(int colors)
         init_pair(9, COLOR_BLACK, COLOR_GREEN);
         statattrs[MS_BRB] = COLOR_PAIR(9);
 
-        init_pair(10, COLOR_YELLOW, COLOR_BLACK);
+        init_pair(10, COLOR_YELLOW, COLOR_BG);
         lstatattrs[MS_IDL] = lstatattrs[MS_AWY] =
                 lstatattrs[MS_LUN] = COLOR_PAIR(10) | A_BOLD;
 
-        init_pair(11, COLOR_BLUE, COLOR_BLACK);
-        attrs[C_GRP] = COLOR_PAIR(11);
+        init_pair(11, COLOR_BLUE, COLOR_BG);
+        C_GRP = COLOR_PAIR(11);
         
-        init_pair(12, COLOR_WHITE, COLOR_BLACK);
-        attrs[C_NORMAL] = lstatattrs[MS_UNK] = COLOR_PAIR(12);
+        init_pair(12, COLOR_WHITE, COLOR_BG);
+        C_NORMAL = lstatattrs[MS_UNK] = COLOR_PAIR(12);
     }
+    msg2_attr = C_MNU;
+    
     cbreak();
     noecho();
     nonl();
@@ -708,7 +960,7 @@ void screen_init(int colors)
     keypad(stdscr, TRUE);
     meta(stdscr, TRUE);
     scrollok(stdscr, TRUE);
-    attrset(attrs[C_NORMAL]);
+    attrset(C_NORMAL);
 
     clear();
     
@@ -718,8 +970,8 @@ void screen_init(int colors)
     move(SLINES - 1, SCOLS - 1);
     refresh();
 
-    w_msg.w = SCOLS * 3 / 4;
-    w_msg.h = 2 * SLINES / 5 - 2;
+    w_msg.w = (int) (SCOLS * scr_rx + 0.5);
+    w_msg.h = (int) (SLINES * scr_ry + 0.5) - 2;
     w_msg.wh = newpad(W_MSG_LBUF, w_msg.w);
 /*    w_msg.wh = newwin(w_msg.h, w_msg.w, 0, 0);*/
     scrollok(w_msg.wh, TRUE);
@@ -749,10 +1001,22 @@ void screen_init(int colors)
     w_back.y = 0;
     
     wvis = 0;
+    clmenu = 0;
 
     pthread_mutex_init(&w_back.lock, NULL);
 
     pthread_mutex_init(&scr_lock, NULL);
     
     fullscreen = 1;
+    while ((q = queue_get(&msg_q)) != NULL) {
+        msg(C_ERR, "%s", q->data);
+        qelem_free(q);
+    }
+}
+
+void screen_done()
+{
+    delwin(w_msg.wh);
+    erase(); refresh();
+    endwin();
 }
