@@ -2,7 +2,7 @@
  *    utf8.h
  *
  *    utf8 support routines
- *    Copyright (C) 2002-2004  George M. Tzoumas
+ *    Copyright (C) 2002-2007  George M. Tzoumas
  *
  *    This program is free software; you can redistribute it and/or modify
  *    it under the terms of the GNU General Public License as published by
@@ -22,6 +22,7 @@
 #ifndef _UTF8_H_
 #define _UTF8_H_
 
+#include <wchar.h>
 #include "../config.h"
 
 int strlen_utf8(char *s, int *len);
@@ -30,14 +31,14 @@ int stroffset(char *s, int n, int len);
 int seqlen(int c);
 int strwidth(char *s);
 int widthoffset(char *s, int width);
+int wstrwidth(const wchar_t *s, size_t n);
 
-#ifndef HAVE_WCSWIDTH
-
-int wcswidth(const wchar_t *s, size_t n);
-#define wcwidth(c) 1
-
+#ifndef HAVE_WCWIDTH
+#define wcwidth(c) (1)
 #endif
 
+#define wcwidth_nl(c) ((c) == '\n' ? 1: wcwidth(c))
+
 #ifndef HAVE_MBSRTOWCS
 
 #define mbsrtowcs(dest, src, len, ps) mbstowcs(dest, *(src), len)
